# Copyright 2021-2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import asyncio
import json
import struct
import sys

import websockets.asyncio.server

import bumble.logging
from bumble.core import (
    BT_HIDP_PROTOCOL_ID,
    BT_HUMAN_INTERFACE_DEVICE_SERVICE,
    BT_L2CAP_PROTOCOL_ID,
    PhysicalTransport,
)
from bumble.device import Device
from bumble.hid import HID_CONTROL_PSM, HID_INTERRUPT_PSM, Message
from bumble.hid import Device as HID_Device
from bumble.sdp import (
    SDP_ADDITIONAL_PROTOCOL_DESCRIPTOR_LIST_ATTRIBUTE_ID,
    SDP_BLUETOOTH_PROFILE_DESCRIPTOR_LIST_ATTRIBUTE_ID,
    SDP_BROWSE_GROUP_LIST_ATTRIBUTE_ID,
    SDP_LANGUAGE_BASE_ATTRIBUTE_ID_LIST_ATTRIBUTE_ID,
    SDP_PROTOCOL_DESCRIPTOR_LIST_ATTRIBUTE_ID,
    SDP_PUBLIC_BROWSE_ROOT,
    SDP_SERVICE_CLASS_ID_LIST_ATTRIBUTE_ID,
    SDP_SERVICE_RECORD_HANDLE_ATTRIBUTE_ID,
    DataElement,
    ServiceAttribute,
)
from bumble.transport import open_transport

# -----------------------------------------------------------------------------
# SDP attributes for Bluetooth HID devices
SDP_HID_SERVICE_NAME_ATTRIBUTE_ID = 0x0100
SDP_HID_SERVICE_DESCRIPTION_ATTRIBUTE_ID = 0x0101
SDP_HID_PROVIDER_NAME_ATTRIBUTE_ID = 0x0102
SDP_HID_DEVICE_RELEASE_NUMBER_ATTRIBUTE_ID = 0x0200  # [DEPRECATED]
SDP_HID_PARSER_VERSION_ATTRIBUTE_ID = 0x0201
SDP_HID_DEVICE_SUBCLASS_ATTRIBUTE_ID = 0x0202
SDP_HID_COUNTRY_CODE_ATTRIBUTE_ID = 0x0203
SDP_HID_VIRTUAL_CABLE_ATTRIBUTE_ID = 0x0204
SDP_HID_RECONNECT_INITIATE_ATTRIBUTE_ID = 0x0205
SDP_HID_DESCRIPTOR_LIST_ATTRIBUTE_ID = 0x0206
SDP_HID_LANGID_BASE_LIST_ATTRIBUTE_ID = 0x0207
SDP_HID_SDP_DISABLE_ATTRIBUTE_ID = 0x0208  # [DEPRECATED]
SDP_HID_BATTERY_POWER_ATTRIBUTE_ID = 0x0209
SDP_HID_REMOTE_WAKE_ATTRIBUTE_ID = 0x020A
SDP_HID_PROFILE_VERSION_ATTRIBUTE_ID = 0x020B  # DEPRECATED]
SDP_HID_SUPERVISION_TIMEOUT_ATTRIBUTE_ID = 0x020C
SDP_HID_NORMALLY_CONNECTABLE_ATTRIBUTE_ID = 0x020D
SDP_HID_BOOT_DEVICE_ATTRIBUTE_ID = 0x020E
SDP_HID_SSR_HOST_MAX_LATENCY_ATTRIBUTE_ID = 0x020F
SDP_HID_SSR_HOST_MIN_TIMEOUT_ATTRIBUTE_ID = 0x0210

# Refer to HID profile specification v1.1.1, "5.3 Service Discovery Protocol (SDP)" for details
# HID SDP attribute values
LANGUAGE = 0x656E  # 0x656E uint16 “en” (English)
ENCODING = 0x6A  # 0x006A uint16 UTF-8 encoding
PRIMARY_LANGUAGE_BASE_ID = 0x100  # 0x0100 uint16 PrimaryLanguageBaseID
VERSION_NUMBER = 0x0101  # 0x0101 uint16 version number (v1.1)
SERVICE_NAME = b'Bumble HID'
SERVICE_DESCRIPTION = b'Bumble'
PROVIDER_NAME = b'Bumble'
HID_PARSER_VERSION = 0x0111  # uint16 0x0111 (v1.1.1)
HID_DEVICE_SUBCLASS = 0xC0  # Combo keyboard/pointing device
HID_COUNTRY_CODE = 0x21  # 0x21 Uint8, USA
HID_VIRTUAL_CABLE = True  # Virtual cable enabled
HID_RECONNECT_INITIATE = True  #  Reconnect initiate enabled
REPORT_DESCRIPTOR_TYPE = 0x22  # 0x22 Type = Report Descriptor
HID_LANGID_BASE_LANGUAGE = 0x0409  # 0x0409 Language = English (United States)
HID_LANGID_BASE_BLUETOOTH_STRING_OFFSET = 0x100  # 0x0100 Default
HID_BATTERY_POWER = True  #  Battery power enabled
HID_REMOTE_WAKE = True  #  Remote wake enabled
HID_SUPERVISION_TIMEOUT = 0xC80  # uint16 0xC80 (2s)
HID_NORMALLY_CONNECTABLE = True  #  Normally connectable enabled
HID_BOOT_DEVICE = True  #  Boot device support enabled
HID_SSR_HOST_MAX_LATENCY = 0x640  # uint16 0x640 (1s)
HID_SSR_HOST_MIN_TIMEOUT = 0xC80  # uint16 0xC80 (2s)
HID_REPORT_MAP = bytes(  # Text String, 50 Octet Report Descriptor
    # pylint: disable=line-too-long
    [
        0x05,
        0x01,  # Usage Page (Generic Desktop Ctrls)
        0x09,
        0x06,  # Usage (Keyboard)
        0xA1,
        0x01,  # Collection (Application)
        0x85,
        0x01,  # . Report ID (1)
        0x05,
        0x07,  # . Usage Page (Kbrd/Keypad)
        0x19,
        0xE0,  # . Usage Minimum (0xE0)
        0x29,
        0xE7,  # . Usage Maximum (0xE7)
        0x15,
        0x00,  # . Logical Minimum (0)
        0x25,
        0x01,  # . Logical Maximum (1)
        0x75,
        0x01,  # . Report Size (1)
        0x95,
        0x08,  # . Report Count (8)
        0x81,
        0x02,  # . Input (Data,Var,Abs,No Wrap,Linear,Preferred State,No Null Position)
        0x95,
        0x01,  # . Report Count (1)
        0x75,
        0x08,  # . Report Size (8)
        0x81,
        0x03,  # . Input (Const,Var,Abs,No Wrap,Linear,Preferred State,No Null Position)
        0x95,
        0x05,  # . Report Count (5)
        0x75,
        0x01,  # . Report Size (1)
        0x05,
        0x08,  # . Usage Page (LEDs)
        0x19,
        0x01,  # . Usage Minimum (Num Lock)
        0x29,
        0x05,  # . Usage Maximum (Kana)
        0x91,
        0x02,  # . Output (Data,Var,Abs,No Wrap,Linear,Preferred State,No Null Position,Non-volatile)
        0x95,
        0x01,  # . Report Count (1)
        0x75,
        0x03,  # . Report Size (3)
        0x91,
        0x03,  # . Output (Const,Var,Abs,No Wrap,Linear,Preferred State,No Null Position,Non-volatile)
        0x95,
        0x06,  # . Report Count (6)
        0x75,
        0x08,  # . Report Size (8)
        0x15,
        0x00,  # . Logical Minimum (0)
        0x25,
        0x65,  # . Logical Maximum (101)
        0x05,
        0x07,  # . Usage Page (Kbrd/Keypad)
        0x19,
        0x00,  # . Usage Minimum (0x00)
        0x29,
        0x65,  # . Usage Maximum (0x65)
        0x81,
        0x00,  # . Input (Data,Array,Abs,No Wrap,Linear,Preferred State,No Null Position)
        0xC0,  # End Collection
        0x05,
        0x01,  # Usage Page (Generic Desktop Ctrls)
        0x09,
        0x02,  # Usage (Mouse)
        0xA1,
        0x01,  # Collection (Application)
        0x85,
        0x02,  # . Report ID (2)
        0x09,
        0x01,  # . Usage (Pointer)
        0xA1,
        0x00,  # . Collection (Physical)
        0x05,
        0x09,  # .   Usage Page (Button)
        0x19,
        0x01,  # .   Usage Minimum (0x01)
        0x29,
        0x03,  # .   Usage Maximum (0x03)
        0x15,
        0x00,  # .   Logical Minimum (0)
        0x25,
        0x01,  # .   Logical Maximum (1)
        0x95,
        0x03,  # .   Report Count (3)
        0x75,
        0x01,  # .   Report Size (1)
        0x81,
        0x02,  # .   Input (Data,Var,Abs,No Wrap,Linear,Preferred State,No Null Position)
        0x95,
        0x01,  # .   Report Count (1)
        0x75,
        0x05,  # .   Report Size (5)
        0x81,
        0x03,  # .   Input (Const,Var,Abs,No Wrap,Linear,Preferred State,No Null Position)
        0x05,
        0x01,  # .   Usage Page (Generic Desktop Ctrls)
        0x09,
        0x30,  # .   Usage (X)
        0x09,
        0x31,  # .   Usage (Y)
        0x15,
        0x81,  # .   Logical Minimum (-127)
        0x25,
        0x7F,  # .   Logical Maximum (127)
        0x75,
        0x08,  # .   Report Size (8)
        0x95,
        0x02,  # .   Report Count (2)
        0x81,
        0x06,  # .   Input (Data,Var,Rel,No Wrap,Linear,Preferred State,No Null Position)
        0xC0,  # . End Collection
        0xC0,  # End Collection
    ]
)


# Default protocol mode set to report protocol
protocol_mode = Message.ProtocolMode.REPORT_PROTOCOL


# -----------------------------------------------------------------------------
def sdp_records():
    service_record_handle = 0x00010002
    return {
        service_record_handle: [
            ServiceAttribute(
                SDP_SERVICE_RECORD_HANDLE_ATTRIBUTE_ID,
                DataElement.unsigned_integer_32(service_record_handle),
            ),
            ServiceAttribute(
                SDP_BROWSE_GROUP_LIST_ATTRIBUTE_ID,
                DataElement.sequence([DataElement.uuid(SDP_PUBLIC_BROWSE_ROOT)]),
            ),
            ServiceAttribute(
                SDP_SERVICE_CLASS_ID_LIST_ATTRIBUTE_ID,
                DataElement.sequence(
                    [DataElement.uuid(BT_HUMAN_INTERFACE_DEVICE_SERVICE)]
                ),
            ),
            ServiceAttribute(
                SDP_PROTOCOL_DESCRIPTOR_LIST_ATTRIBUTE_ID,
                DataElement.sequence(
                    [
                        DataElement.sequence(
                            [
                                DataElement.uuid(BT_L2CAP_PROTOCOL_ID),
                                DataElement.unsigned_integer_16(HID_CONTROL_PSM),
                            ]
                        ),
                        DataElement.sequence(
                            [
                                DataElement.uuid(BT_HIDP_PROTOCOL_ID),
                            ]
                        ),
                    ]
                ),
            ),
            ServiceAttribute(
                SDP_LANGUAGE_BASE_ATTRIBUTE_ID_LIST_ATTRIBUTE_ID,
                DataElement.sequence(
                    [
                        DataElement.unsigned_integer_16(LANGUAGE),
                        DataElement.unsigned_integer_16(ENCODING),
                        DataElement.unsigned_integer_16(PRIMARY_LANGUAGE_BASE_ID),
                    ]
                ),
            ),
            ServiceAttribute(
                SDP_BLUETOOTH_PROFILE_DESCRIPTOR_LIST_ATTRIBUTE_ID,
                DataElement.sequence(
                    [
                        DataElement.sequence(
                            [
                                DataElement.uuid(BT_HUMAN_INTERFACE_DEVICE_SERVICE),
                                DataElement.unsigned_integer_16(VERSION_NUMBER),
                            ]
                        ),
                    ]
                ),
            ),
            ServiceAttribute(
                SDP_ADDITIONAL_PROTOCOL_DESCRIPTOR_LIST_ATTRIBUTE_ID,
                DataElement.sequence(
                    [
                        DataElement.sequence(
                            [
                                DataElement.sequence(
                                    [
                                        DataElement.uuid(BT_L2CAP_PROTOCOL_ID),
                                        DataElement.unsigned_integer_16(
                                            HID_INTERRUPT_PSM
                                        ),
                                    ]
                                ),
                                DataElement.sequence(
                                    [
                                        DataElement.uuid(BT_HIDP_PROTOCOL_ID),
                                    ]
                                ),
                            ]
                        ),
                    ]
                ),
            ),
            ServiceAttribute(
                SDP_HID_SERVICE_NAME_ATTRIBUTE_ID,
                DataElement(DataElement.TEXT_STRING, SERVICE_NAME),
            ),
            ServiceAttribute(
                SDP_HID_SERVICE_DESCRIPTION_ATTRIBUTE_ID,
                DataElement(DataElement.TEXT_STRING, SERVICE_DESCRIPTION),
            ),
            ServiceAttribute(
                SDP_HID_PROVIDER_NAME_ATTRIBUTE_ID,
                DataElement(DataElement.TEXT_STRING, PROVIDER_NAME),
            ),
            ServiceAttribute(
                SDP_HID_PARSER_VERSION_ATTRIBUTE_ID,
                DataElement.unsigned_integer_32(HID_PARSER_VERSION),
            ),
            ServiceAttribute(
                SDP_HID_DEVICE_SUBCLASS_ATTRIBUTE_ID,
                DataElement.unsigned_integer_32(HID_DEVICE_SUBCLASS),
            ),
            ServiceAttribute(
                SDP_HID_COUNTRY_CODE_ATTRIBUTE_ID,
                DataElement.unsigned_integer_32(HID_COUNTRY_CODE),
            ),
            ServiceAttribute(
                SDP_HID_VIRTUAL_CABLE_ATTRIBUTE_ID,
                DataElement.boolean(HID_VIRTUAL_CABLE),
            ),
            ServiceAttribute(
                SDP_HID_RECONNECT_INITIATE_ATTRIBUTE_ID,
                DataElement.boolean(HID_RECONNECT_INITIATE),
            ),
            ServiceAttribute(
                SDP_HID_DESCRIPTOR_LIST_ATTRIBUTE_ID,
                DataElement.sequence(
                    [
                        DataElement.sequence(
                            [
                                DataElement.unsigned_integer_16(REPORT_DESCRIPTOR_TYPE),
                                DataElement(DataElement.TEXT_STRING, HID_REPORT_MAP),
                            ]
                        ),
                    ]
                ),
            ),
            ServiceAttribute(
                SDP_HID_LANGID_BASE_LIST_ATTRIBUTE_ID,
                DataElement.sequence(
                    [
                        DataElement.sequence(
                            [
                                DataElement.unsigned_integer_16(
                                    HID_LANGID_BASE_LANGUAGE
                                ),
                                DataElement.unsigned_integer_16(
                                    HID_LANGID_BASE_BLUETOOTH_STRING_OFFSET
                                ),
                            ]
                        ),
                    ]
                ),
            ),
            ServiceAttribute(
                SDP_HID_BATTERY_POWER_ATTRIBUTE_ID,
                DataElement.boolean(HID_BATTERY_POWER),
            ),
            ServiceAttribute(
                SDP_HID_REMOTE_WAKE_ATTRIBUTE_ID,
                DataElement.boolean(HID_REMOTE_WAKE),
            ),
            ServiceAttribute(
                SDP_HID_SUPERVISION_TIMEOUT_ATTRIBUTE_ID,
                DataElement.unsigned_integer_16(HID_SUPERVISION_TIMEOUT),
            ),
            ServiceAttribute(
                SDP_HID_NORMALLY_CONNECTABLE_ATTRIBUTE_ID,
                DataElement.boolean(HID_NORMALLY_CONNECTABLE),
            ),
            ServiceAttribute(
                SDP_HID_BOOT_DEVICE_ATTRIBUTE_ID,
                DataElement.boolean(HID_BOOT_DEVICE),
            ),
            ServiceAttribute(
                SDP_HID_SSR_HOST_MAX_LATENCY_ATTRIBUTE_ID,
                DataElement.unsigned_integer_16(HID_SSR_HOST_MAX_LATENCY),
            ),
            ServiceAttribute(
                SDP_HID_SSR_HOST_MIN_TIMEOUT_ATTRIBUTE_ID,
                DataElement.unsigned_integer_16(HID_SSR_HOST_MIN_TIMEOUT),
            ),
        ]
    }


# -----------------------------------------------------------------------------
async def get_stream_reader(pipe) -> asyncio.StreamReader:
    loop = asyncio.get_event_loop()
    reader = asyncio.StreamReader(loop=loop)
    protocol = asyncio.StreamReaderProtocol(reader)
    await loop.connect_read_pipe(lambda: protocol, pipe)
    return reader


class DeviceData:
    def __init__(self) -> None:
        self.keyboardData = bytearray(
            [0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00]
        )
        self.mouseData = bytearray([0x02, 0x00, 0x00, 0x00])


# Device's live data - Mouse and Keyboard will be stored in this
deviceData = DeviceData()


# -----------------------------------------------------------------------------
async def keyboard_device(hid_device: HID_Device):
    # Start a Websocket server to receive events from a web page
    async def serve(websocket: websockets.asyncio.server.ServerConnection):
        global deviceData
        while True:
            try:
                message = await websocket.recv()
                print('Received: ', str(message))
                parsed = json.loads(message)
                message_type = parsed['type']
                if message_type == 'keydown':
                    # Only deal with keys a to z for now
                    key = parsed['key']
                    if len(key) == 1:
                        code = ord(key)
                        if ord('a') <= code <= ord('z'):
                            hid_code = 0x04 + code - ord('a')
                            deviceData.keyboardData = bytearray(
                                [
                                    0x01,
                                    0x00,
                                    0x00,
                                    hid_code,
                                    0x00,
                                    0x00,
                                    0x00,
                                    0x00,
                                    0x00,
                                ]
                            )
                            hid_device.send_data(deviceData.keyboardData)
                elif message_type == 'keyup':
                    deviceData.keyboardData = bytearray(
                        [0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00]
                    )
                    hid_device.send_data(deviceData.keyboardData)
                elif message_type == "mousemove":
                    # logical min and max values
                    log_min = -127
                    log_max = 127
                    x = parsed['x']
                    y = parsed['y']
                    # limiting x and y values within logical max and min range
                    x = max(log_min, min(log_max, x))
                    y = max(log_min, min(log_max, y))
                    deviceData.mouseData = bytearray([0x02, 0x00]) + struct.pack(
                        ">bb", x, y
                    )
                    hid_device.send_data(deviceData.mouseData)
            except websockets.exceptions.ConnectionClosedOK:
                pass

    # pylint: disable-next=no-member
    await websockets.asyncio.server.serve(serve, 'localhost', 8989)
    await asyncio.get_event_loop().create_future()


# -----------------------------------------------------------------------------
async def main() -> None:
    if len(sys.argv) < 3:
        print(
            'Usage: python run_hid_device.py <device-config> <transport-spec> <command>'
            '  where <command> is one of:\n'
            '  test-mode (run with menu enabled for testing)\n'
            '  web (run a keyboard with keypress input from a web page, '
            'see keyboard.html'
        )
        print('example: python run_hid_device.py hid_keyboard.json usb:0 web')
        print('example: python run_hid_device.py hid_keyboard.json usb:0 test-mode')

        return

    async def handle_virtual_cable_unplug():
        hid_host_bd_addr = str(hid_device.remote_device_bd_address)
        await hid_device.disconnect_interrupt_channel()
        await hid_device.disconnect_control_channel()
        await device.keystore.delete(hid_host_bd_addr)  # type: ignore
        connection = hid_device.connection
        if connection is not None:
            await connection.disconnect()

    def on_hid_data_cb(pdu: bytes):
        print(f'Received Data, PDU: {pdu.hex()}')

    def on_get_report_cb(
        report_id: int, report_type: int, buffer_size: int
    ) -> HID_Device.GetSetStatus:
        retValue = hid_device.GetSetStatus()
        print(
            "GET_REPORT report_id: "
            + str(report_id)
            + "report_type: "
            + str(report_type)
            + "buffer_size:"
            + str(buffer_size)
        )
        if report_type == Message.ReportType.INPUT_REPORT:
            if report_id == 1:
                retValue.data = deviceData.keyboardData[1:]
                retValue.status = hid_device.GetSetReturn.SUCCESS
            elif report_id == 2:
                retValue.data = deviceData.mouseData[1:]
                retValue.status = hid_device.GetSetReturn.SUCCESS
            else:
                retValue.status = hid_device.GetSetReturn.REPORT_ID_NOT_FOUND

            if buffer_size:
                data_len = buffer_size - 1
                retValue.data = retValue.data[:data_len]
        elif report_type == Message.ReportType.OUTPUT_REPORT:
            # This sample app has nothing to do with the report received, to enable PTS
            # testing, we will return single byte random data.
            retValue.data = bytearray([0x11])
            retValue.status = hid_device.GetSetReturn.SUCCESS
        elif report_type == Message.ReportType.FEATURE_REPORT:
            retValue.status = hid_device.GetSetReturn.ERR_INVALID_PARAMETER
        elif report_type == Message.ReportType.OTHER_REPORT:
            if report_id == 3:
                retValue.status = hid_device.GetSetReturn.REPORT_ID_NOT_FOUND
        else:
            retValue.status = hid_device.GetSetReturn.FAILURE

        return retValue

    def on_set_report_cb(
        report_id: int, report_type: int, report_size: int, data: bytes
    ) -> HID_Device.GetSetStatus:
        print(
            "SET_REPORT report_id: "
            + str(report_id)
            + "report_type: "
            + str(report_type)
            + "report_size "
            + str(report_size)
            + "data:"
            + str(data)
        )
        if report_type == Message.ReportType.FEATURE_REPORT:
            status = HID_Device.GetSetReturn.ERR_INVALID_PARAMETER
        elif report_type == Message.ReportType.INPUT_REPORT:
            if report_id == 1 and report_size != len(deviceData.keyboardData):
                status = HID_Device.GetSetReturn.ERR_INVALID_PARAMETER
            elif report_id == 2 and report_size != len(deviceData.mouseData):
                status = HID_Device.GetSetReturn.ERR_INVALID_PARAMETER
            elif report_id == 3:
                status = HID_Device.GetSetReturn.REPORT_ID_NOT_FOUND
            else:
                status = HID_Device.GetSetReturn.SUCCESS
        else:
            status = HID_Device.GetSetReturn.SUCCESS

        return HID_Device.GetSetStatus(status=status)

    def on_get_protocol_cb() -> HID_Device.GetSetStatus:
        return HID_Device.GetSetStatus(
            data=bytes([protocol_mode]),
            status=hid_device.GetSetReturn.SUCCESS,
        )

    def on_set_protocol_cb(protocol: int) -> HID_Device.GetSetStatus:
        # We do not support SET_PROTOCOL.
        print(f"SET_PROTOCOL report_id: {protocol}")
        return HID_Device.GetSetStatus(
            status=hid_device.GetSetReturn.ERR_UNSUPPORTED_REQUEST
        )

    def on_virtual_cable_unplug_cb():
        print('Received Virtual Cable Unplug')
        asyncio.create_task(handle_virtual_cable_unplug())

    print('<<< connecting to HCI...')
    async with await open_transport(sys.argv[2]) as hci_transport:
        print('<<< connected')

        # Create a device
        device = Device.from_config_file_with_hci(
            sys.argv[1], hci_transport.source, hci_transport.sink
        )
        device.classic_enabled = True

        # Create and register HID device
        hid_device = HID_Device(device)

        # Register for  call backs
        hid_device.on('interrupt_data', on_hid_data_cb)

        hid_device.register_get_report_cb(on_get_report_cb)
        hid_device.register_set_report_cb(on_set_report_cb)
        hid_device.register_get_protocol_cb(on_get_protocol_cb)
        hid_device.register_set_protocol_cb(on_set_protocol_cb)

        # Register for virtual cable unplug call back
        hid_device.on('virtual_cable_unplug', on_virtual_cable_unplug_cb)

        # Setup the SDP to advertise HID Device service
        device.sdp_service_records = sdp_records()

        # Start the controller
        await device.power_on()

        # Start being discoverable and connectable
        await device.set_discoverable(True)
        await device.set_connectable(True)

        async def menu():
            reader = await get_stream_reader(sys.stdin)
            while True:
                print(
                    "\n************************ HID Device Menu *****************************\n"
                )
                print(" 1. Connect Control Channel")
                print(" 2. Connect Interrupt Channel")
                print(" 3. Disconnect Control Channel")
                print(" 4. Disconnect Interrupt Channel")
                print(" 5. Send Report on Interrupt Channel")
                print(" 6. Virtual Cable Unplug")
                print(" 7. Disconnect device")
                print(" 8. Delete Bonding")
                print(" 9. Re-connect to device")
                print("10. Exit ")
                print("\nEnter your choice : \n")

                choice = await reader.readline()
                choice = choice.decode('utf-8').strip()

                if choice == '1':
                    await hid_device.connect_control_channel()

                elif choice == '2':
                    await hid_device.connect_interrupt_channel()

                elif choice == '3':
                    await hid_device.disconnect_control_channel()

                elif choice == '4':
                    await hid_device.disconnect_interrupt_channel()

                elif choice == '5':
                    print(" 1. Report ID 0x01")
                    print(" 2. Report ID 0x02")
                    print(" 3. Invalid Report ID")

                    choice1 = await reader.readline()
                    choice1 = choice1.decode('utf-8').strip()

                    if choice1 == '1':
                        data = bytearray(
                            [0x01, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00]
                        )
                        hid_device.send_data(data)
                        data = bytearray(
                            [0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00]
                        )
                        hid_device.send_data(data)

                    elif choice1 == '2':
                        data = bytearray([0x02, 0x00, 0x00, 0xF6])
                        hid_device.send_data(data)
                        data = bytearray([0x02, 0x00, 0x00, 0x00])
                        hid_device.send_data(data)

                    elif choice1 == '3':
                        data = bytearray([0x00, 0x00, 0x00, 0x00])
                        hid_device.send_data(data)
                        data = bytearray([0x00, 0x00, 0x00, 0x00])
                        hid_device.send_data(data)

                    else:
                        print('Incorrect option selected')

                elif choice == '6':
                    hid_device.virtual_cable_unplug()
                    try:
                        hid_host_bd_addr = str(hid_device.remote_device_bd_address)
                        await device.keystore.delete(hid_host_bd_addr)
                    except KeyError:
                        print('Device not found or Device already unpaired.')

                elif choice == '7':
                    connection = hid_device.connection
                    if connection is not None:
                        await connection.disconnect()
                    else:
                        print("Already disconnected from device")

                elif choice == '8':
                    try:
                        hid_host_bd_addr = str(hid_device.remote_device_bd_address)
                        await device.keystore.delete(hid_host_bd_addr)
                    except KeyError:
                        print('Device NOT found or Device already unpaired.')

                elif choice == '9':
                    hid_host_bd_addr = str(hid_device.remote_device_bd_address)
                    connection = await device.connect(
                        hid_host_bd_addr, transport=PhysicalTransport.BR_EDR
                    )
                    await connection.authenticate()
                    await connection.encrypt()

                elif choice == '10':
                    sys.exit("Exit successful")

                else:
                    print("Invalid option selected.")

        if (len(sys.argv) > 3) and (sys.argv[3] == 'test-mode'):
            # Test mode for PTS/Unit testing
            await menu()
        else:
            # default option is using keyboard.html (web)
            print("Executing in Web mode")
            await keyboard_device(hid_device)

        await hci_transport.source.terminated


# -----------------------------------------------------------------------------
bumble.logging.setup_basic_logging('DEBUG')
asyncio.run(main())



# Copyright (c) 2012 Giorgos Verigakis <verigak@gmail.com>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from functools import partial


class ColorError(ValueError):
    """Error raised when a color spec is invalid."""


# ANSI color names. There is also a "default"
COLORS = ('black', 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white')

# ANSI style names
STYLES = (
    'none',
    'bold',
    'faint',
    'italic',
    'underline',
    'blink',
    'blink2',
    'negative',
    'concealed',
    'crossed',
)


ColorSpec = str | int


def _join(*values: ColorSpec) -> str:
    return ';'.join(str(v) for v in values)


def _color_code(spec: ColorSpec, base: int) -> str:
    if isinstance(spec, str):
        spec = spec.strip().lower()

    if spec == 'default':
        return _join(base + 9)
    elif spec in COLORS:
        return _join(base + COLORS.index(spec))
    elif isinstance(spec, int) and 0 <= spec <= 255:
        return _join(base + 8, 5, spec)
    else:
        raise ColorError(f'Invalid color spec "{spec}"')


def color(
    s: str,
    fg: ColorSpec | None = None,
    bg: ColorSpec | None = None,
    style: str | None = None,
) -> str:
    codes: list[ColorSpec] = []

    if fg:
        codes.append(_color_code(fg, 30))
    if bg:
        codes.append(_color_code(bg, 40))
    if style:
        for style_part in style.split('+'):
            if style_part in STYLES:
                codes.append(STYLES.index(style_part))
            else:
                raise ColorError(f'Invalid style "{style_part}"')

    if codes:
        return f'\x1b[{_join(*codes)}m{s}\x1b[0m'
    else:
        return s


# Foreground color shortcuts
black = partial(color, fg='black')
red = partial(color, fg='red')
green = partial(color, fg='green')
yellow = partial(color, fg='yellow')
blue = partial(color, fg='blue')
magenta = partial(color, fg='magenta')
cyan = partial(color, fg='cyan')
white = partial(color, fg='white')

# Style shortcuts
bold = partial(color, style='bold')
none = partial(color, style='none')
faint = partial(color, style='faint')
italic = partial(color, style='italic')
underline = partial(color, style='underline')
blink = partial(color, style='blink')
blink2 = partial(color, style='blink2')
negative = partial(color, style='negative')
concealed = partial(color, style='concealed')
crossed = partial(color, style='crossed')



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
from __future__ import annotations

import asyncio
import logging
from dataclasses import dataclass, field
from typing import Any

from bumble import att, gatt, gatt_adapters, gatt_client, utils
from bumble.core import InvalidArgumentError, InvalidStateError
from bumble.device import Connection, Device
from bumble.hci import Address


# -----------------------------------------------------------------------------
# Constants
# -----------------------------------------------------------------------------
class ErrorCode(utils.OpenIntEnum):
    '''See Hearing Access Service 2.4. Attribute Profile error codes.'''

    INVALID_OPCODE = 0x80
    WRITE_NAME_NOT_ALLOWED = 0x81
    PRESET_SYNCHRONIZATION_NOT_SUPPORTED = 0x82
    PRESET_OPERATION_NOT_POSSIBLE = 0x83
    INVALID_PARAMETERS_LENGTH = 0x84


class HearingAidType(utils.OpenIntEnum):
    '''See Hearing Access Service 3.1. Hearing Aid Features.'''

    BINAURAL_HEARING_AID = 0b00
    MONAURAL_HEARING_AID = 0b01
    BANDED_HEARING_AID = 0b10


class PresetSynchronizationSupport(utils.OpenIntEnum):
    '''See Hearing Access Service 3.1. Hearing Aid Features.'''

    PRESET_SYNCHRONIZATION_IS_NOT_SUPPORTED = 0b0
    PRESET_SYNCHRONIZATION_IS_SUPPORTED = 0b1


class IndependentPresets(utils.OpenIntEnum):
    '''See Hearing Access Service 3.1. Hearing Aid Features.'''

    IDENTICAL_PRESET_RECORD = 0b0
    DIFFERENT_PRESET_RECORD = 0b1


class DynamicPresets(utils.OpenIntEnum):
    '''See Hearing Access Service 3.1. Hearing Aid Features.'''

    PRESET_RECORDS_DOES_NOT_CHANGE = 0b0
    PRESET_RECORDS_MAY_CHANGE = 0b1


class WritablePresetsSupport(utils.OpenIntEnum):
    '''See Hearing Access Service 3.1. Hearing Aid Features.'''

    WRITABLE_PRESET_RECORDS_NOT_SUPPORTED = 0b0
    WRITABLE_PRESET_RECORDS_SUPPORTED = 0b1


class HearingAidPresetControlPointOpcode(utils.OpenIntEnum):
    '''See Hearing Access Service 3.3.1 Hearing Aid Preset Control Point operation requirements.'''

    # fmt: off
    READ_PRESETS_REQUEST                     = 0x01
    READ_PRESET_RESPONSE                     = 0x02
    PRESET_CHANGED                           = 0x03
    WRITE_PRESET_NAME                        = 0x04
    SET_ACTIVE_PRESET                        = 0x05
    SET_NEXT_PRESET                          = 0x06
    SET_PREVIOUS_PRESET                      = 0x07
    SET_ACTIVE_PRESET_SYNCHRONIZED_LOCALLY   = 0x08
    SET_NEXT_PRESET_SYNCHRONIZED_LOCALLY     = 0x09
    SET_PREVIOUS_PRESET_SYNCHRONIZED_LOCALLY = 0x0A


@dataclass
class HearingAidFeatures:
    '''See Hearing Access Service 3.1. Hearing Aid Features.'''

    hearing_aid_type: HearingAidType
    preset_synchronization_support: PresetSynchronizationSupport
    independent_presets: IndependentPresets
    dynamic_presets: DynamicPresets
    writable_presets_support: WritablePresetsSupport

    def __bytes__(self) -> bytes:
        return bytes(
            [
                (self.hearing_aid_type << 0)
                | (self.preset_synchronization_support << 2)
                | (self.independent_presets << 3)
                | (self.dynamic_presets << 4)
                | (self.writable_presets_support << 5)
            ]
        )


def HearingAidFeatures_from_bytes(data: int) -> HearingAidFeatures:
    return HearingAidFeatures(
        HearingAidType(data & 0b11),
        PresetSynchronizationSupport(data >> 2 & 0b1),
        IndependentPresets(data >> 3 & 0b1),
        DynamicPresets(data >> 4 & 0b1),
        WritablePresetsSupport(data >> 5 & 0b1),
    )


@dataclass
class PresetChangedOperation:
    '''See Hearing Access Service 3.2.2.2. Preset Changed operation.'''

    class ChangeId(utils.OpenIntEnum):
        # fmt: off
        GENERIC_UPDATE            = 0x00
        PRESET_RECORD_DELETED     = 0x01
        PRESET_RECORD_AVAILABLE   = 0x02
        PRESET_RECORD_UNAVAILABLE = 0x03

    @dataclass
    class Generic:
        prev_index: int
        preset_record: PresetRecord

        def __bytes__(self) -> bytes:
            return bytes([self.prev_index]) + bytes(self.preset_record)

    change_id: ChangeId
    additional_parameters: Generic | int

    def to_bytes(self, is_last: bool) -> bytes:
        if isinstance(self.additional_parameters, PresetChangedOperation.Generic):
            additional_parameters_bytes = bytes(self.additional_parameters)
        else:
            additional_parameters_bytes = bytes([self.additional_parameters])

        return (
            bytes(
                [
                    HearingAidPresetControlPointOpcode.PRESET_CHANGED,
                    self.change_id,
                    is_last,
                ]
            )
            + additional_parameters_bytes
        )


class PresetChangedOperationDeleted(PresetChangedOperation):
    def __init__(self, index) -> None:
        self.change_id = PresetChangedOperation.ChangeId.PRESET_RECORD_DELETED
        self.additional_parameters = index


class PresetChangedOperationAvailable(PresetChangedOperation):
    def __init__(self, index) -> None:
        self.change_id = PresetChangedOperation.ChangeId.PRESET_RECORD_AVAILABLE
        self.additional_parameters = index


class PresetChangedOperationUnavailable(PresetChangedOperation):
    def __init__(self, index) -> None:
        self.change_id = PresetChangedOperation.ChangeId.PRESET_RECORD_UNAVAILABLE
        self.additional_parameters = index


@dataclass
class PresetRecord:
    '''See Hearing Access Service 2.8. Preset record.'''

    @dataclass
    class Property:
        class Writable(utils.OpenIntEnum):
            CANNOT_BE_WRITTEN = 0b0
            CAN_BE_WRITTEN = 0b1

        class IsAvailable(utils.OpenIntEnum):
            IS_UNAVAILABLE = 0b0
            IS_AVAILABLE = 0b1

        writable: Writable = Writable.CAN_BE_WRITTEN
        is_available: IsAvailable = IsAvailable.IS_AVAILABLE

        def __bytes__(self) -> bytes:
            return bytes([self.writable | (self.is_available << 1)])

    index: int
    name: str
    properties: Property = field(default_factory=Property)

    def __bytes__(self) -> bytes:
        return bytes([self.index]) + bytes(self.properties) + self.name.encode('utf-8')

    def is_available(self) -> bool:
        return (
            self.properties.is_available
            == PresetRecord.Property.IsAvailable.IS_AVAILABLE
        )


# -----------------------------------------------------------------------------
# Server
# -----------------------------------------------------------------------------
class HearingAccessService(gatt.TemplateService):
    UUID = gatt.GATT_HEARING_ACCESS_SERVICE

    hearing_aid_features_characteristic: gatt.Characteristic[bytes]
    hearing_aid_preset_control_point: gatt.Characteristic[bytes]
    active_preset_index_characteristic: gatt.Characteristic[bytes]
    active_preset_index: int
    active_preset_index_per_device: dict[Address, int]

    device: Device

    server_features: HearingAidFeatures
    preset_records: dict[int, PresetRecord]  # key is the preset index
    read_presets_request_in_progress: bool

    other_server_in_binaural_set: HearingAccessService | None = None

    preset_changed_operations_history_per_device: dict[
        Address, list[PresetChangedOperation]
    ]

    # Keep an updated list of connected client to send notification to
    currently_connected_clients: set[Connection]

    def __init__(
        self, device: Device, features: HearingAidFeatures, presets: list[PresetRecord]
    ) -> None:
        self.active_preset_index_per_device = {}
        self.read_presets_request_in_progress = False
        self.preset_changed_operations_history_per_device = {}
        self.currently_connected_clients = set()

        self.device = device
        self.server_features = features
        if len(presets) < 1:
            raise InvalidArgumentError(f'Invalid presets: {presets}')

        self.preset_records = {}
        for p in presets:
            if len(p.name.encode()) < 1 or len(p.name.encode()) > 40:
                raise InvalidArgumentError(f'Invalid name: {p.name}')

            self.preset_records[p.index] = p

        # associate the lowest index as the current active preset at startup
        self.active_preset_index = sorted(self.preset_records.keys())[0]

        @device.on(device.EVENT_CONNECTION)
        def on_connection(connection: Connection) -> None:
            @connection.on(connection.EVENT_DISCONNECTION)
            def on_disconnection(_reason) -> None:
                self.currently_connected_clients.discard(connection)

            @connection.on(connection.EVENT_CONNECTION_ATT_MTU_UPDATE)
            def on_mtu_update(*_: Any) -> None:
                self.on_incoming_connection(connection)

            @connection.on(connection.EVENT_CONNECTION_ENCRYPTION_CHANGE)
            def on_encryption_change(*_: Any) -> None:
                self.on_incoming_connection(connection)

            @connection.on(connection.EVENT_PAIRING)
            def on_pairing(*_: Any) -> None:
                self.on_incoming_connection(connection)

            self.on_incoming_connection(connection)

        self.hearing_aid_features_characteristic = gatt.Characteristic(
            uuid=gatt.GATT_HEARING_AID_FEATURES_CHARACTERISTIC,
            properties=gatt.Characteristic.Properties.READ,
            permissions=gatt.Characteristic.Permissions.READ_REQUIRES_ENCRYPTION,
            value=bytes(self.server_features),
        )
        self.hearing_aid_preset_control_point = gatt.Characteristic(
            uuid=gatt.GATT_HEARING_AID_PRESET_CONTROL_POINT_CHARACTERISTIC,
            properties=(
                gatt.Characteristic.Properties.WRITE
                | gatt.Characteristic.Properties.INDICATE
            ),
            permissions=gatt.Characteristic.Permissions.WRITE_REQUIRES_ENCRYPTION,
            value=gatt.CharacteristicValue(
                write=self._on_write_hearing_aid_preset_control_point
            ),
        )
        self.active_preset_index_characteristic = gatt.Characteristic(
            uuid=gatt.GATT_ACTIVE_PRESET_INDEX_CHARACTERISTIC,
            properties=(
                gatt.Characteristic.Properties.READ
                | gatt.Characteristic.Properties.NOTIFY
            ),
            permissions=gatt.Characteristic.Permissions.READ_REQUIRES_ENCRYPTION,
            value=gatt.CharacteristicValue(read=self._on_read_active_preset_index),
        )

        super().__init__(
            [
                self.hearing_aid_features_characteristic,
                self.hearing_aid_preset_control_point,
                self.active_preset_index_characteristic,
            ]
        )

    def on_incoming_connection(self, connection: Connection):
        '''Setup initial operations to handle a remote bonded HAP device'''
        # TODO Should we filter on HAP device only ?

        if not connection.is_encrypted:
            logging.debug(f'HAS: {connection.peer_address} is not encrypted')
            return

        if not connection.peer_resolvable_address:
            logging.debug(f'HAS: {connection.peer_address} is not paired')
            return

        if connection.att_mtu < 49:
            logging.debug(
                f'HAS: {connection.peer_address} invalid MTU={connection.att_mtu}'
            )
            return

        if connection.peer_address in self.currently_connected_clients:
            logging.debug(
                f'HAS: Already connected to {connection.peer_address} nothing to do'
            )
            return

        self.currently_connected_clients.add(connection)
        if (
            connection.peer_address
            not in self.preset_changed_operations_history_per_device
        ):
            self.preset_changed_operations_history_per_device[
                connection.peer_address
            ] = []
            return

        async def on_connection_async() -> None:
            # Send all the PresetChangedOperation that occur when not connected
            await self._preset_changed_operation(connection)
            # Update the active preset index if needed
            await self.notify_active_preset_for_connection(connection)

        connection.cancel_on_disconnection(on_connection_async())

    def _on_read_active_preset_index(self, connection: Connection) -> bytes:
        del connection  # Unused
        return bytes([self.active_preset_index])

    # TODO this need to be triggered when device is unbonded
    def on_forget(self, addr: Address) -> None:
        self.preset_changed_operations_history_per_device.pop(addr)

    async def _on_write_hearing_aid_preset_control_point(
        self, connection: Connection, value: bytes
    ):
        opcode = HearingAidPresetControlPointOpcode(value[0])
        handler = getattr(self, '_on_' + opcode.name.lower())
        await handler(connection, value)

    async def _on_read_presets_request(self, connection: Connection, value: bytes):
        if connection.att_mtu < 49:  # 2.5. GATT sub-procedure requirements
            logging.warning(f'HAS require MTU >= 49: {connection}')

        if self.read_presets_request_in_progress:
            raise att.ATT_Error(att.ErrorCode.PROCEDURE_ALREADY_IN_PROGRESS)
        self.read_presets_request_in_progress = True

        start_index = value[1]
        if start_index == 0x00:
            raise att.ATT_Error(att.ErrorCode.OUT_OF_RANGE)

        num_presets = value[2]
        if num_presets == 0x00:
            raise att.ATT_Error(att.ErrorCode.OUT_OF_RANGE)

        # Sending `num_presets` presets ordered by increasing index field, starting from start_index
        presets = [
            self.preset_records[key]
            for key in sorted(self.preset_records.keys())
            if self.preset_records[key].index >= start_index
        ][:num_presets]
        if len(presets) == 0:
            raise att.ATT_Error(att.ErrorCode.OUT_OF_RANGE)

        utils.AsyncRunner.spawn(self._read_preset_response(connection, presets))

    async def _read_preset_response(
        self, connection: Connection, presets: list[PresetRecord]
    ):
        # If the ATT bearer is terminated before all notifications or indications are
        # sent, then the server shall consider the Read Presets Request operation
        # aborted and shall not either continue or restart the operation when the client
        # reconnects.
        try:
            for i, preset in enumerate(presets):
                await connection.device.indicate_subscriber(
                    connection,
                    self.hearing_aid_preset_control_point,
                    value=bytes(
                        [
                            HearingAidPresetControlPointOpcode.READ_PRESET_RESPONSE,
                            i == len(presets) - 1,
                        ]
                    )
                    + bytes(preset),
                )

        finally:
            # indicate_subscriber can raise a TimeoutError, we need to gracefully terminate the operation
            self.read_presets_request_in_progress = False

    async def generic_update(self, op: PresetChangedOperation) -> None:
        '''Server API to perform a generic update. It is the responsibility of the caller to modify the preset_records to match the PresetChangedOperation being sent'''
        await self._notify_preset_operations(op)

    async def delete_preset(self, index: int) -> None:
        '''Server API to delete a preset. It should not be the current active preset'''

        if index == self.active_preset_index:
            raise InvalidStateError('Cannot delete active preset')

        del self.preset_records[index]
        await self._notify_preset_operations(PresetChangedOperationDeleted(index))

    async def available_preset(self, index: int) -> None:
        '''Server API to make a preset available'''

        preset = self.preset_records[index]
        preset.properties.is_available = PresetRecord.Property.IsAvailable.IS_AVAILABLE
        await self._notify_preset_operations(PresetChangedOperationAvailable(index))

    async def unavailable_preset(self, index: int) -> None:
        '''Server API to make a preset unavailable. It should not be the current active preset'''

        if index == self.active_preset_index:
            raise InvalidStateError('Cannot set active preset as unavailable')

        preset = self.preset_records[index]
        preset.properties.is_available = (
            PresetRecord.Property.IsAvailable.IS_UNAVAILABLE
        )
        await self._notify_preset_operations(PresetChangedOperationUnavailable(index))

    async def _preset_changed_operation(self, connection: Connection) -> None:
        '''Send all PresetChangedOperation saved for a given connection'''
        op_list = self.preset_changed_operations_history_per_device.get(
            connection.peer_address, []
        )

        # Notification will be sent in index order
        def get_op_index(op: PresetChangedOperation) -> int:
            if isinstance(op.additional_parameters, PresetChangedOperation.Generic):
                return op.additional_parameters.prev_index
            return op.additional_parameters

        op_list.sort(key=get_op_index)
        # If the ATT bearer is terminated before all notifications or indications are
        # sent, then the server shall consider the Preset Changed operation aborted and
        # shall continue the operation when the client reconnects.
        while op_list:
            try:
                await connection.device.indicate_subscriber(
                    connection,
                    self.hearing_aid_preset_control_point,
                    value=op_list[0].to_bytes(len(op_list) == 1),
                    force=True,  # TODO GATT notification subscription should be persistent
                )
                # Remove item once sent, and keep the non sent item in the list
                op_list.pop(0)
            except TimeoutError:
                break

    async def _notify_preset_operations(self, op: PresetChangedOperation) -> None:
        for history_list in self.preset_changed_operations_history_per_device.values():
            history_list.append(op)

        for connection in self.currently_connected_clients:
            await self._preset_changed_operation(connection)

    async def _on_write_preset_name(self, connection: Connection, value: bytes):
        del connection  # Unused

        if self.read_presets_request_in_progress:
            raise att.ATT_Error(att.ErrorCode.PROCEDURE_ALREADY_IN_PROGRESS)

        index = value[1]
        preset = self.preset_records.get(index, None)
        if (
            not preset
            or preset.properties.writable
            == PresetRecord.Property.Writable.CANNOT_BE_WRITTEN
        ):
            raise att.ATT_Error(ErrorCode.WRITE_NAME_NOT_ALLOWED)

        name = value[2:].decode('utf-8')
        if not name or len(name) > 40:
            raise att.ATT_Error(ErrorCode.INVALID_PARAMETERS_LENGTH)

        preset.name = name

        await self.generic_update(
            PresetChangedOperation(
                PresetChangedOperation.ChangeId.GENERIC_UPDATE,
                PresetChangedOperation.Generic(index, preset),
            )
        )

    async def notify_active_preset_for_connection(self, connection: Connection) -> None:
        if (
            self.active_preset_index_per_device.get(connection.peer_address, 0x00)
            == self.active_preset_index
        ):
            # Nothing to do, peer is already updated
            return

        await connection.device.notify_subscriber(
            connection,
            attribute=self.active_preset_index_characteristic,
            value=bytes([self.active_preset_index]),
        )
        self.active_preset_index_per_device[connection.peer_address] = (
            self.active_preset_index
        )

    async def notify_active_preset(self) -> None:
        for connection in self.currently_connected_clients:
            await self.notify_active_preset_for_connection(connection)

    async def set_active_preset(self, value: bytes) -> None:
        index = value[1]
        preset = self.preset_records.get(index, None)
        if (
            not preset
            or preset.properties.is_available
            != PresetRecord.Property.IsAvailable.IS_AVAILABLE
        ):
            raise att.ATT_Error(ErrorCode.PRESET_OPERATION_NOT_POSSIBLE)

        if index == self.active_preset_index:
            # Already at correct value
            return

        self.active_preset_index = index
        await self.notify_active_preset()

    async def _on_set_active_preset(self, connection: Connection, value: bytes):
        del connection  # Unused
        await self.set_active_preset(value)

    async def set_next_or_previous_preset(self, is_previous: bool) -> None:
        '''Set the next or the previous preset as active'''

        if self.active_preset_index == 0x00:
            raise att.ATT_Error(ErrorCode.PRESET_OPERATION_NOT_POSSIBLE)

        presets = sorted(
            [
                record
                for record in self.preset_records.values()
                if record.is_available()
            ],
            key=lambda record: record.index,
        )
        current_preset = self.preset_records[self.active_preset_index]
        current_preset_pos = presets.index(current_preset)
        if is_previous:
            new_preset = presets[(current_preset_pos - 1) % len(presets)]
        else:
            new_preset = presets[(current_preset_pos + 1) % len(presets)]

        if current_preset == new_preset:  # If no other preset are available
            raise att.ATT_Error(ErrorCode.PRESET_OPERATION_NOT_POSSIBLE)

        self.active_preset_index = new_preset.index
        await self.notify_active_preset()

    async def _on_set_next_preset(self, connection: Connection, value: bytes) -> None:
        del connection, value  # Unused.
        await self.set_next_or_previous_preset(False)

    async def _on_set_previous_preset(
        self, connection: Connection, value: bytes
    ) -> None:
        del connection, value  # Unused.
        await self.set_next_or_previous_preset(True)

    async def _on_set_active_preset_synchronized_locally(
        self, connection: Connection, value: bytes
    ):
        del connection  # Unused.
        if (
            self.server_features.preset_synchronization_support
            == PresetSynchronizationSupport.PRESET_SYNCHRONIZATION_IS_NOT_SUPPORTED
        ):
            raise att.ATT_Error(ErrorCode.PRESET_SYNCHRONIZATION_NOT_SUPPORTED)
        await self.set_active_preset(value)
        if self.other_server_in_binaural_set:
            await self.other_server_in_binaural_set.set_active_preset(value)

    async def _on_set_next_preset_synchronized_locally(
        self, connection: Connection, value: bytes
    ):
        del connection, value  # Unused.
        if (
            self.server_features.preset_synchronization_support
            == PresetSynchronizationSupport.PRESET_SYNCHRONIZATION_IS_NOT_SUPPORTED
        ):
            raise att.ATT_Error(ErrorCode.PRESET_SYNCHRONIZATION_NOT_SUPPORTED)
        await self.set_next_or_previous_preset(False)
        if self.other_server_in_binaural_set:
            await self.other_server_in_binaural_set.set_next_or_previous_preset(False)

    async def _on_set_previous_preset_synchronized_locally(
        self, connection: Connection, value: bytes
    ):
        del connection, value  # Unused.
        if (
            self.server_features.preset_synchronization_support
            == PresetSynchronizationSupport.PRESET_SYNCHRONIZATION_IS_NOT_SUPPORTED
        ):
            raise att.ATT_Error(ErrorCode.PRESET_SYNCHRONIZATION_NOT_SUPPORTED)
        await self.set_next_or_previous_preset(True)
        if self.other_server_in_binaural_set:
            await self.other_server_in_binaural_set.set_next_or_previous_preset(True)


# -----------------------------------------------------------------------------
# Client
# -----------------------------------------------------------------------------
class HearingAccessServiceProxy(gatt_client.ProfileServiceProxy):
    SERVICE_CLASS = HearingAccessService

    hearing_aid_preset_control_point: gatt_client.CharacteristicProxy
    preset_control_point_indications: asyncio.Queue[bytes]
    active_preset_index_notification: asyncio.Queue[bytes]

    def __init__(self, service_proxy: gatt_client.ServiceProxy) -> None:
        self.service_proxy = service_proxy
        self.preset_control_point_indications = asyncio.Queue()
        self.active_preset_index_notification = asyncio.Queue()

        self.server_features = gatt_adapters.PackedCharacteristicProxyAdapter(
            service_proxy.get_characteristics_by_uuid(
                gatt.GATT_HEARING_AID_FEATURES_CHARACTERISTIC
            )[0],
            'B',
        )

        self.hearing_aid_preset_control_point = (
            service_proxy.get_characteristics_by_uuid(
                gatt.GATT_HEARING_AID_PRESET_CONTROL_POINT_CHARACTERISTIC
            )[0]
        )

        self.active_preset_index = gatt_adapters.PackedCharacteristicProxyAdapter(
            service_proxy.get_characteristics_by_uuid(
                gatt.GATT_ACTIVE_PRESET_INDEX_CHARACTERISTIC
            )[0],
            'B',
        )

    async def setup_subscription(self) -> None:
        await self.hearing_aid_preset_control_point.subscribe(
            self.preset_control_point_indications.put_nowait,
            prefer_notify=False,
        )

        await self.active_preset_index.subscribe(
            self.active_preset_index_notification.put_nowait
        )



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import struct
from dataclasses import dataclass

from bumble import utils
from bumble.att import ATT_Error
from bumble.device import Connection
from bumble.gatt import (
    GATT_AUDIO_LOCATION_CHARACTERISTIC,
    GATT_AUDIO_OUTPUT_DESCRIPTION_CHARACTERISTIC,
    GATT_VOLUME_OFFSET_CONTROL_POINT_CHARACTERISTIC,
    GATT_VOLUME_OFFSET_CONTROL_SERVICE,
    GATT_VOLUME_OFFSET_STATE_CHARACTERISTIC,
    Characteristic,
    CharacteristicValue,
    TemplateService,
)
from bumble.gatt_adapters import (
    DelegatedCharacteristicProxyAdapter,
    SerializableCharacteristicProxyAdapter,
    UTF8CharacteristicProxyAdapter,
)
from bumble.gatt_client import ProfileServiceProxy, ServiceProxy
from bumble.profiles.bap import AudioLocation

# -----------------------------------------------------------------------------
# Constants
# -----------------------------------------------------------------------------

MIN_VOLUME_OFFSET = -255
MAX_VOLUME_OFFSET = 255
CHANGE_COUNTER_MAX_VALUE = 0xFF


class SetVolumeOffsetOpCode(utils.OpenIntEnum):
    SET_VOLUME_OFFSET = 0x01


class ErrorCode(utils.OpenIntEnum):
    """
    See Volume Offset Control Service 1.6. Application error codes.
    """

    INVALID_CHANGE_COUNTER = 0x80
    OPCODE_NOT_SUPPORTED = 0x81
    VALUE_OUT_OF_RANGE = 0x82


# -----------------------------------------------------------------------------
@dataclass
class VolumeOffsetState:
    volume_offset: int = 0
    change_counter: int = 0
    attribute: Characteristic | None = None

    def __bytes__(self) -> bytes:
        return struct.pack('<hB', self.volume_offset, self.change_counter)

    @classmethod
    def from_bytes(cls, data: bytes):
        volume_offset, change_counter = struct.unpack('<hB', data)
        return cls(volume_offset, change_counter)

    def increment_change_counter(self) -> None:
        self.change_counter = (self.change_counter + 1) % (CHANGE_COUNTER_MAX_VALUE + 1)

    async def notify_subscribers_via_connection(self, connection: Connection) -> None:
        assert self.attribute is not None
        await connection.device.notify_subscribers(attribute=self.attribute)

    def on_read(self, _connection: Connection) -> bytes:
        return bytes(self)


@dataclass
class VocsAudioLocation:
    audio_location: AudioLocation = AudioLocation.NOT_ALLOWED
    attribute: Characteristic | None = None

    def __bytes__(self) -> bytes:
        return struct.pack('<I', self.audio_location)

    @classmethod
    def from_bytes(cls, data: bytes):
        audio_location = AudioLocation(struct.unpack('<I', data)[0])
        return cls(audio_location)

    def on_read(self, _connection: Connection) -> bytes:
        return bytes(self)

    async def on_write(self, connection: Connection, value: bytes) -> None:
        assert self.attribute

        self.audio_location = AudioLocation(int.from_bytes(value, 'little'))
        await connection.device.notify_subscribers(attribute=self.attribute)


@dataclass
class VolumeOffsetControlPoint:
    volume_offset_state: VolumeOffsetState

    async def on_write(self, connection: Connection, value: bytes) -> None:
        opcode = value[0]
        if opcode != SetVolumeOffsetOpCode.SET_VOLUME_OFFSET:
            raise ATT_Error(ErrorCode.OPCODE_NOT_SUPPORTED)

        change_counter, volume_offset = struct.unpack('<Bh', value[1:])
        await self._set_volume_offset(connection, change_counter, volume_offset)

    async def _set_volume_offset(
        self,
        connection: Connection,
        change_counter_operand: int,
        volume_offset_operand: int,
    ) -> None:
        change_counter = self.volume_offset_state.change_counter

        if change_counter != change_counter_operand:
            raise ATT_Error(ErrorCode.INVALID_CHANGE_COUNTER)

        if not MIN_VOLUME_OFFSET <= volume_offset_operand <= MAX_VOLUME_OFFSET:
            raise ATT_Error(ErrorCode.VALUE_OUT_OF_RANGE)

        self.volume_offset_state.volume_offset = volume_offset_operand
        self.volume_offset_state.increment_change_counter()
        await self.volume_offset_state.notify_subscribers_via_connection(connection)


@dataclass
class AudioOutputDescription:
    audio_output_description: str = ''
    attribute: Characteristic | None = None

    @classmethod
    def from_bytes(cls, data: bytes):
        return cls(audio_output_description=data.decode('utf-8'))

    def __bytes__(self) -> bytes:
        return self.audio_output_description.encode('utf-8')

    def on_read(self, _connection: Connection) -> bytes:
        return bytes(self)

    async def on_write(self, connection: Connection, value: bytes) -> None:
        assert self.attribute

        self.audio_output_description = value.decode('utf-8')
        await connection.device.notify_subscribers(attribute=self.attribute)


# -----------------------------------------------------------------------------
class VolumeOffsetControlService(TemplateService):
    UUID = GATT_VOLUME_OFFSET_CONTROL_SERVICE

    def __init__(
        self,
        volume_offset_state: VolumeOffsetState | None = None,
        audio_location: VocsAudioLocation | None = None,
        audio_output_description: AudioOutputDescription | None = None,
    ) -> None:
        self.volume_offset_state = (
            VolumeOffsetState() if volume_offset_state is None else volume_offset_state
        )

        self.audio_location = (
            VocsAudioLocation() if audio_location is None else audio_location
        )

        self.audio_output_description = (
            AudioOutputDescription()
            if audio_output_description is None
            else audio_output_description
        )

        self.volume_offset_control_point: VolumeOffsetControlPoint = (
            VolumeOffsetControlPoint(self.volume_offset_state)
        )

        self.volume_offset_state_characteristic: Characteristic[bytes] = Characteristic(
            uuid=GATT_VOLUME_OFFSET_STATE_CHARACTERISTIC,
            properties=(
                Characteristic.Properties.READ | Characteristic.Properties.NOTIFY
            ),
            permissions=Characteristic.Permissions.READ_REQUIRES_ENCRYPTION,
            value=CharacteristicValue(read=self.volume_offset_state.on_read),
        )

        self.audio_location_characteristic: Characteristic[bytes] = Characteristic(
            uuid=GATT_AUDIO_LOCATION_CHARACTERISTIC,
            properties=(
                Characteristic.Properties.READ
                | Characteristic.Properties.NOTIFY
                | Characteristic.Properties.WRITE_WITHOUT_RESPONSE
            ),
            permissions=(
                Characteristic.Permissions.READ_REQUIRES_ENCRYPTION
                | Characteristic.Permissions.WRITE_REQUIRES_ENCRYPTION
            ),
            value=CharacteristicValue(
                read=self.audio_location.on_read,
                write=self.audio_location.on_write,
            ),
        )
        self.audio_location.attribute = self.audio_location_characteristic

        self.volume_offset_control_point_characteristic: Characteristic[bytes] = (
            Characteristic(
                uuid=GATT_VOLUME_OFFSET_CONTROL_POINT_CHARACTERISTIC,
                properties=Characteristic.Properties.WRITE,
                permissions=Characteristic.Permissions.WRITE_REQUIRES_ENCRYPTION,
                value=CharacteristicValue(
                    write=self.volume_offset_control_point.on_write
                ),
            )
        )

        self.audio_output_description_characteristic: Characteristic[bytes] = (
            Characteristic(
                uuid=GATT_AUDIO_OUTPUT_DESCRIPTION_CHARACTERISTIC,
                properties=(
                    Characteristic.Properties.READ
                    | Characteristic.Properties.NOTIFY
                    | Characteristic.Properties.WRITE_WITHOUT_RESPONSE
                ),
                permissions=(
                    Characteristic.Permissions.READ_REQUIRES_ENCRYPTION
                    | Characteristic.Permissions.WRITE_REQUIRES_ENCRYPTION
                ),
                value=CharacteristicValue(
                    read=self.audio_output_description.on_read,
                    write=self.audio_output_description.on_write,
                ),
            )
        )
        self.audio_output_description.attribute = (
            self.audio_output_description_characteristic
        )

        super().__init__(
            characteristics=[
                self.volume_offset_state_characteristic,  # type: ignore
                self.audio_location_characteristic,  # type: ignore
                self.volume_offset_control_point_characteristic,  # type: ignore
                self.audio_output_description_characteristic,  # type: ignore
            ],
            primary=False,
        )


# -----------------------------------------------------------------------------
# Client
# -----------------------------------------------------------------------------
class VolumeOffsetControlServiceProxy(ProfileServiceProxy):
    SERVICE_CLASS = VolumeOffsetControlService

    def __init__(self, service_proxy: ServiceProxy) -> None:
        self.service_proxy = service_proxy

        self.volume_offset_state = SerializableCharacteristicProxyAdapter(
            service_proxy.get_required_characteristic_by_uuid(
                GATT_VOLUME_OFFSET_STATE_CHARACTERISTIC
            ),
            VolumeOffsetState,
        )

        self.audio_location = DelegatedCharacteristicProxyAdapter(
            service_proxy.get_required_characteristic_by_uuid(
                GATT_AUDIO_LOCATION_CHARACTERISTIC
            ),
            encode=lambda value: bytes([int(value)]),
            decode=lambda data: AudioLocation(data[0]),
        )

        self.volume_offset_control_point = (
            service_proxy.get_required_characteristic_by_uuid(
                GATT_VOLUME_OFFSET_CONTROL_POINT_CHARACTERISTIC
            )
        )

        self.audio_output_description = UTF8CharacteristicProxyAdapter(
            service_proxy.get_required_characteristic_by_uuid(
                GATT_AUDIO_OUTPUT_DESCRIPTION_CHARACTERISTIC
            )
        )



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import contextlib
import functools
import inspect
import logging
from collections.abc import Generator, MutableMapping
from typing import Any

import grpc
from google.protobuf.message import Message  # pytype: disable=pyi-error

from bumble.device import Device
from bumble.hci import Address, AddressType

ADDRESS_TYPES: dict[str, AddressType] = {
    "public": Address.PUBLIC_DEVICE_ADDRESS,
    "random": Address.RANDOM_DEVICE_ADDRESS,
    "public_identity": Address.PUBLIC_IDENTITY_ADDRESS,
    "random_static_identity": Address.RANDOM_IDENTITY_ADDRESS,
}


def address_from_request(request: Message, field: str | None) -> Address:
    if field is None:
        return Address.ANY
    return Address(bytes(reversed(getattr(request, field))), ADDRESS_TYPES[field])


class BumbleServerLoggerAdapter(logging.LoggerAdapter):  # type: ignore
    """Formats logs from the PandoraClient."""

    def process(
        self, msg: str, kwargs: MutableMapping[str, Any]
    ) -> tuple[str, MutableMapping[str, Any]]:
        assert self.extra
        service_name = self.extra['service_name']
        assert isinstance(service_name, str)
        device = self.extra['device']
        assert isinstance(device, Device)
        addr_bytes = bytes(
            reversed(bytes(device.public_address))
        )  # pytype: disable=attribute-error
        addr = ':'.join([f'{x:02X}' for x in addr_bytes[4:]])
        return (f'[bumble.{service_name}:{addr}] {msg}', kwargs)


@contextlib.contextmanager
def exception_to_rpc_error(
    context: grpc.ServicerContext,
) -> Generator[None, None, None]:
    try:
        yield None
    except NotImplementedError as e:
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)  # type: ignore
        context.set_details(str(e))  # type: ignore
    except ValueError as e:
        context.set_code(grpc.StatusCode.INVALID_ARGUMENT)  # type: ignore
        context.set_details(str(e))  # type: ignore
    except RuntimeError as e:
        context.set_code(grpc.StatusCode.ABORTED)  # type: ignore
        context.set_details(str(e))  # type: ignore


# Decorate an RPC servicer method with a wrapper that transform exceptions to gRPC errors.
def rpc(func: Any) -> Any:
    @functools.wraps(func)
    async def asyncgen_wrapper(
        self: Any, request: Any, context: grpc.ServicerContext
    ) -> Any:
        with exception_to_rpc_error(context):
            async for v in func(self, request, context):
                yield v

    @functools.wraps(func)
    async def async_wrapper(
        self: Any, request: Any, context: grpc.ServicerContext
    ) -> Any:
        with exception_to_rpc_error(context):
            return await func(self, request, context)

    @functools.wraps(func)
    def gen_wrapper(self: Any, request: Any, context: grpc.ServicerContext) -> Any:
        with exception_to_rpc_error(context):
            yield from func(self, request, context)

    @functools.wraps(func)
    def wrapper(self: Any, request: Any, context: grpc.ServicerContext) -> Any:
        with exception_to_rpc_error(context):
            return func(self, request, context)

    if inspect.isasyncgenfunction(func):
        return asyncgen_wrapper

    if inspect.iscoroutinefunction(func):
        return async_wrapper

    if inspect.isgenerator(func):
        return gen_wrapper

    return wrapper



# Copyright 2021-2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import asyncio
import itertools
import logging
import os
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from bumble.core import PhysicalTransport, ProtocolError
from bumble.device import Peer
from bumble.gatt import Characteristic, Service
from bumble.hci import Role
from bumble.pairing import PairingConfig, PairingDelegate
from bumble.smp import (
    ErrorCode,
    OobContext,
    OobLegacyContext,
)

from .test_utils import TwoDevices, async_barrier

# -----------------------------------------------------------------------------
# Logging
# -----------------------------------------------------------------------------
logger = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
@pytest.mark.asyncio
async def test_self_connection():
    two_devices = TwoDevices()
    await two_devices.setup_connection()


# -----------------------------------------------------------------------------
@pytest.mark.asyncio
async def test_self_disconnection():
    two_devices = TwoDevices()
    await two_devices.setup_connection()
    await two_devices.connections[0].disconnect()
    await async_barrier()
    assert not two_devices.connections

    two_devices = TwoDevices()
    await two_devices.setup_connection()
    await two_devices.connections[1].disconnect()
    await async_barrier()
    assert not two_devices.connections


# -----------------------------------------------------------------------------
@pytest.mark.asyncio
@pytest.mark.parametrize(
    'responder_role,',
    (Role.CENTRAL, Role.PERIPHERAL),
)
async def test_self_classic_connection(responder_role):
    two_devices = TwoDevices()

    # Enable Classic connections
    two_devices.devices[0].classic_enabled = True
    two_devices.devices[1].classic_enabled = True

    # Start
    for dev in two_devices.devices:
        await dev.power_on()

    # Connect the two devices
    await asyncio.gather(
        two_devices.devices[0].connect(
            two_devices.devices[1].public_address, transport=PhysicalTransport.BR_EDR
        ),
        two_devices.devices[1].accept(
            two_devices.devices[0].public_address, responder_role
        ),
    )

    # Check the post conditions
    assert two_devices.connections[0] is not None
    assert two_devices.connections[1] is not None

    # Check the role
    assert two_devices.connections[0].role != responder_role
    assert two_devices.connections[1].role == responder_role

    # Role switch
    await two_devices.connections[0].switch_role(responder_role)

    # Check the role
    assert two_devices.connections[0].role == responder_role
    assert two_devices.connections[1].role != responder_role

    await two_devices.connections[0].disconnect()


# -----------------------------------------------------------------------------
@pytest.mark.asyncio
async def test_self_gatt():
    # Create two devices, each with a controller, attached to the same link
    two_devices = TwoDevices()

    # Add some GATT characteristics to device 1
    c1 = Characteristic(
        '3A143AD7-D4A7-436B-97D6-5B62C315E833',
        Characteristic.Properties.READ,
        Characteristic.READABLE,
        bytes([1, 2, 3]),
    )
    c2 = Characteristic(
        '9557CCE2-DB37-46EB-94C4-50AE5B9CB0F8',
        Characteristic.Properties.READ | Characteristic.Properties.WRITE,
        Characteristic.READABLE | Characteristic.WRITEABLE,
        bytes([4, 5, 6]),
    )
    c3 = Characteristic(
        '84FC1A2E-C52D-4A2D-B8C3-8855BAB86638',
        Characteristic.Properties.READ
        | Characteristic.Properties.WRITE_WITHOUT_RESPONSE,
        Characteristic.READABLE | Characteristic.WRITEABLE,
        bytes([7, 8, 9]),
    )
    c4 = Characteristic(
        '84FC1A2E-C52D-4A2D-B8C3-8855BAB86638',
        Characteristic.Properties.READ
        | Characteristic.Properties.NOTIFY
        | Characteristic.Properties.INDICATE,
        Characteristic.READABLE,
        bytes([1, 1, 1]),
    )

    s1 = Service('8140E247-04F0-42C1-BC34-534C344DAFCA', [c1, c2, c3])
    s2 = Service('97210A0F-1875-4D05-9E5D-326EB171257A', [c4])
    s3 = Service('1853', [])
    s4 = Service('3A12C182-14E2-4FE0-8C5B-65D7C569F9DB', [], included_services=[s2, s3])
    two_devices.devices[1].add_services([s1, s2, s4])

    # Connect the two devices
    await two_devices.setup_connection()
    peer = Peer(two_devices.connections[0])

    bogus_uuid = 'A0AA6007-0B48-4BBE-80AC-0DE9AAF541EA'
    result = await peer.discover_services([bogus_uuid])
    assert result == []
    services = peer.get_services_by_uuid(bogus_uuid)
    assert len(services) == 0

    result = await peer.discover_service(s1.uuid)
    assert len(result) == 1
    services = peer.get_services_by_uuid(s1.uuid)
    assert len(services) == 1
    s = services[0]
    assert services[0].uuid == s1.uuid

    result = await peer.discover_characteristics([c1.uuid], s)
    assert len(result) == 1
    characteristics = peer.get_characteristics_by_uuid(c1.uuid)
    assert len(characteristics) == 1
    c = characteristics[0]
    assert c.uuid == c1.uuid
    result = await peer.read_value(c)
    assert result is not None
    assert result == c1.value

    result = await peer.discover_service(s4.uuid)
    assert len(result) == 1
    result = await peer.discover_included_services(result[0])
    assert len(result) == 2
    # Service UUID is only present when the UUID is 16-bit Bluetooth UUID
    assert bytes(result[1].uuid) == bytes(s3.uuid)


# -----------------------------------------------------------------------------
@pytest.mark.asyncio
async def test_self_gatt_long_read():
    # Create two devices, each with a controller, attached to the same link
    two_devices = TwoDevices()

    # Add some GATT characteristics to device 1
    characteristics = [
        Characteristic(
            f'3A143AD7-D4A7-436B-97D6-5B62C315{i:04X}',
            Characteristic.Properties.READ,
            Characteristic.READABLE,
            bytes([x & 255 for x in range(i)]),
        )
        for i in range(0, 513)
    ]

    service = Service('8140E247-04F0-42C1-BC34-534C344DAFCA', characteristics)
    two_devices.devices[1].add_service(service)

    # Connect the two devices
    await two_devices.setup_connection()
    peer = Peer(two_devices.connections[0])

    result = await peer.discover_service(service.uuid)
    assert len(result) == 1
    found_service = result[0]
    found_characteristics = await found_service.discover_characteristics()
    assert len(found_characteristics) == 513
    for i, characteristic in enumerate(found_characteristics):
        value = await characteristic.read_value()
        assert value == characteristics[i].value


# -----------------------------------------------------------------------------
async def _test_self_smp_with_configs(pairing_config1, pairing_config2):
    # Create two devices, each with a controller, attached to the same link
    two_devices = TwoDevices()

    # Connect the two devices
    await two_devices.setup_connection()
    connection = two_devices.connections[0]
    assert not connection.is_encrypted

    # Attach pairing listeners
    two_devices.connections[0].on(
        'pairing', lambda keys: two_devices.on_paired(0, keys)
    )
    two_devices.connections[1].on(
        'pairing', lambda keys: two_devices.on_paired(1, keys)
    )

    # Set up the pairing configs
    if pairing_config1:
        two_devices.devices[0].pairing_config_factory = (
            lambda connection: pairing_config1
        )
    if pairing_config2:
        two_devices.devices[1].pairing_config_factory = (
            lambda connection: pairing_config2
        )

    # Pair
    await two_devices.devices[0].pair(connection)
    assert connection.is_encrypted
    assert await two_devices.paired[0] is not None
    assert await two_devices.paired[1] is not None


# -----------------------------------------------------------------------------
IO_CAP = [
    PairingDelegate.IoCapability.NO_OUTPUT_NO_INPUT,
    PairingDelegate.IoCapability.KEYBOARD_INPUT_ONLY,
    PairingDelegate.IoCapability.DISPLAY_OUTPUT_ONLY,
    PairingDelegate.IoCapability.DISPLAY_OUTPUT_AND_YES_NO_INPUT,
    PairingDelegate.IoCapability.DISPLAY_OUTPUT_AND_KEYBOARD_INPUT,
]
SC = [False, True]
MITM = [False, True]
# Key distribution is a 4-bit bitmask
KEY_DIST = range(16)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    'io_caps, sc, mitm, key_dist',
    itertools.chain(
        itertools.product([IO_CAP], SC, MITM, [15]),
        itertools.product(
            [[PairingDelegate.IoCapability.DISPLAY_OUTPUT_AND_KEYBOARD_INPUT]],
            SC,
            MITM,
            KEY_DIST,
        ),
    ),
)
async def test_self_smp(io_caps, sc, mitm, key_dist):
    class Delegate(PairingDelegate):
        def __init__(
            self,
            name,
            io_capability,
            local_initiator_key_distribution,
            local_responder_key_distribution,
        ):
            super().__init__(
                io_capability,
                local_initiator_key_distribution,
                local_responder_key_distribution,
            )
            self.name = name
            self.reset()

        def reset(self):
            self.peer_delegate = None
            self.number = asyncio.get_running_loop().create_future()

        # pylint: disable-next=unused-argument
        async def compare_numbers(self, number, digits):
            if self.peer_delegate is None:
                logger.warning(f'[{self.name}] no peer delegate')
                return False
            await self.display_number(number, digits=6)
            logger.debug(f'[{self.name}] waiting for peer number')
            peer_number = await self.peer_delegate.number
            logger.debug(f'[{self.name}] comparing numbers: {number} and {peer_number}')
            return number == peer_number

        async def get_number(self):
            if self.peer_delegate is None:
                logger.warning(f'[{self.name}] no peer delegate')
                return 0
            else:
                if (
                    self.io_capability
                    == PairingDelegate.IoCapability.KEYBOARD_INPUT_ONLY
                    and self.peer_delegate.io_capability
                    == PairingDelegate.IoCapability.KEYBOARD_INPUT_ONLY
                ):
                    peer_number = 6789
                else:
                    logger.debug(f'[{self.name}] waiting for peer number')
                    peer_number = await self.peer_delegate.number
                logger.debug(f'[{self.name}] returning number: {peer_number}')
                return peer_number

        async def display_number(self, number, digits):
            logger.debug(f'[{self.name}] displaying number: {number}')
            self.number.set_result(number)

        def __str__(self):
            return f'Delegate(name={self.name}, io_capability={self.io_capability})'

    pairing_config_sets = [('Initiator', [None]), ('Responder', [None])]
    for pairing_config_set in pairing_config_sets:
        for io_cap in io_caps:
            delegate = Delegate(pairing_config_set[0], io_cap, key_dist, key_dist)
            pairing_config_set[1].append(PairingConfig(sc, mitm, True, delegate))

    for pairing_config1 in pairing_config_sets[0][1]:
        for pairing_config2 in pairing_config_sets[1][1]:
            logger.info(
                f'########## self_smp with {pairing_config1} and {pairing_config2}'
            )
            if pairing_config1:
                pairing_config1.delegate.reset()
            if pairing_config2:
                pairing_config2.delegate.reset()
            if pairing_config1 and pairing_config2:
                pairing_config1.delegate.peer_delegate = pairing_config2.delegate
                pairing_config2.delegate.peer_delegate = pairing_config1.delegate

            await _test_self_smp_with_configs(pairing_config1, pairing_config2)


# -----------------------------------------------------------------------------
@pytest.mark.asyncio
async def test_self_smp_reject():
    class RejectingDelegate(PairingDelegate):
        def __init__(self):
            super().__init__(PairingDelegate.IoCapability.NO_OUTPUT_NO_INPUT)

        async def accept(self):
            return False

    rejecting_pairing_config = PairingConfig(delegate=RejectingDelegate())
    paired = False
    try:
        await _test_self_smp_with_configs(None, rejecting_pairing_config)
        paired = True
    except ProtocolError as error:
        assert error.error_code == ErrorCode.PAIRING_NOT_SUPPORTED

    assert not paired


# -----------------------------------------------------------------------------
@pytest.mark.asyncio
async def test_self_smp_wrong_pin():
    class WrongPinDelegate(PairingDelegate):
        def __init__(self):
            super().__init__(
                PairingDelegate.IoCapability.DISPLAY_OUTPUT_AND_KEYBOARD_INPUT
            )

        async def compare_numbers(self, number, digits):
            return False

    wrong_pin_pairing_config = PairingConfig(mitm=True, delegate=WrongPinDelegate())
    paired = False
    try:
        await _test_self_smp_with_configs(
            wrong_pin_pairing_config, wrong_pin_pairing_config
        )
        paired = True
    except ProtocolError as error:
        assert error.error_code == ErrorCode.CONFIRM_VALUE_FAILED

    assert not paired


# -----------------------------------------------------------------------------
@pytest.mark.asyncio
async def test_self_smp_over_classic():
    # Create two devices, each with a controller, attached to the same link
    two_devices = TwoDevices()

    # Enable Classic connections
    two_devices.devices[0].classic_enabled = True
    two_devices.devices[1].classic_enabled = True

    # Connect the two devices
    for dev in two_devices.devices:
        await dev.power_on()

    await asyncio.gather(
        two_devices.devices[0].connect(
            two_devices.devices[1].public_address, transport=PhysicalTransport.BR_EDR
        ),
        two_devices.devices[1].accept(two_devices.devices[0].public_address),
    )

    # Check the post conditions
    assert two_devices.connections[0] is not None
    assert two_devices.connections[1] is not None

    # Mock connection
    # TODO: Implement Classic SSP and encryption in link relayer
    LINK_KEY = bytes.fromhex('287ad379dca402530a39f1f43047b835')
    two_devices.devices[0].get_link_key = AsyncMock(return_value=LINK_KEY)
    two_devices.devices[1].get_link_key = AsyncMock(return_value=LINK_KEY)
    two_devices.connections[0].encryption = 1
    two_devices.connections[1].encryption = 1

    two_devices.connections[0].on(
        'pairing', lambda keys: two_devices.on_paired(0, keys)
    )
    two_devices.connections[1].on(
        'pairing', lambda keys: two_devices.on_paired(1, keys)
    )

    # Mock SMP
    with patch('bumble.smp.Session', spec=True) as MockSmpSession:
        MockSmpSession.send_pairing_confirm_command = MagicMock()
        MockSmpSession.send_pairing_dhkey_check_command = MagicMock()
        MockSmpSession.send_public_key_command = MagicMock()
        MockSmpSession.send_pairing_random_command = MagicMock()

        # Start CTKD
        await two_devices.connections[0].pair()
        await asyncio.gather(*two_devices.paired)

        # Phase 2 commands should not be invoked
        MockSmpSession.send_pairing_confirm_command.assert_not_called()
        MockSmpSession.send_pairing_dhkey_check_command.assert_not_called()
        MockSmpSession.send_public_key_command.assert_not_called()
        MockSmpSession.send_pairing_random_command.assert_not_called()

    for i in range(2):
        assert (
            await two_devices.devices[i].keystore.get(
                str(two_devices.connections[i].peer_address)
            )
        ).link_key


# -----------------------------------------------------------------------------
@pytest.mark.asyncio
async def test_self_smp_public_address():
    pairing_config = PairingConfig(
        mitm=True,
        sc=True,
        bonding=True,
        identity_address_type=PairingConfig.AddressType.PUBLIC,
        delegate=PairingDelegate(
            PairingDelegate.IoCapability.DISPLAY_OUTPUT_AND_YES_NO_INPUT,
            PairingDelegate.KeyDistribution.DISTRIBUTE_ENCRYPTION_KEY
            | PairingDelegate.KeyDistribution.DISTRIBUTE_IDENTITY_KEY
            | PairingDelegate.KeyDistribution.DISTRIBUTE_SIGNING_KEY
            | PairingDelegate.KeyDistribution.DISTRIBUTE_LINK_KEY,
        ),
    )

    await _test_self_smp_with_configs(pairing_config, pairing_config)


# -----------------------------------------------------------------------------
@pytest.mark.asyncio
async def test_self_smp_oob_sc():
    oob_context_1 = OobContext()
    oob_context_2 = OobContext()

    pairing_config_1 = PairingConfig(
        mitm=True,
        sc=True,
        bonding=True,
        oob=PairingConfig.OobConfig(oob_context_1, oob_context_2.share(), None),
    )

    pairing_config_2 = PairingConfig(
        mitm=True,
        sc=True,
        bonding=True,
        oob=PairingConfig.OobConfig(oob_context_2, oob_context_1.share(), None),
    )

    await _test_self_smp_with_configs(pairing_config_1, pairing_config_2)

    pairing_config_3 = PairingConfig(
        mitm=True,
        sc=True,
        bonding=True,
        oob=PairingConfig.OobConfig(oob_context_2, None, None),
    )

    await _test_self_smp_with_configs(pairing_config_1, pairing_config_3)
    await _test_self_smp_with_configs(pairing_config_3, pairing_config_1)

    pairing_config_4 = PairingConfig(
        mitm=True,
        sc=True,
        bonding=True,
        oob=PairingConfig.OobConfig(oob_context_2, oob_context_2.share(), None),
    )

    with pytest.raises(ProtocolError) as error:
        await _test_self_smp_with_configs(pairing_config_1, pairing_config_4)
    assert error.value.error_code == ErrorCode.CONFIRM_VALUE_FAILED

    with pytest.raises(ProtocolError):
        await _test_self_smp_with_configs(pairing_config_4, pairing_config_1)
    assert error.value.error_code == ErrorCode.CONFIRM_VALUE_FAILED


# -----------------------------------------------------------------------------
@pytest.mark.asyncio
async def test_self_smp_oob_legacy():
    legacy_context = OobLegacyContext()

    pairing_config_1 = PairingConfig(
        mitm=True,
        sc=False,
        bonding=True,
        oob=PairingConfig.OobConfig(None, None, legacy_context),
    )

    pairing_config_2 = PairingConfig(
        mitm=True,
        sc=True,
        bonding=True,
        oob=PairingConfig.OobConfig(OobContext(), None, legacy_context),
    )

    await _test_self_smp_with_configs(pairing_config_1, pairing_config_2)
    await _test_self_smp_with_configs(pairing_config_2, pairing_config_1)


# -----------------------------------------------------------------------------
async def run_test_self():
    await test_self_connection()
    await test_self_disconnection()
    await test_self_gatt()
    await test_self_gatt_long_read()
    await test_self_smp()
    await test_self_smp_reject()
    await test_self_smp_wrong_pin()
    await test_self_smp_over_classic()
    await test_self_smp_public_address()
    await test_self_smp_oob_sc()
    await test_self_smp_oob_legacy()


# -----------------------------------------------------------------------------
if __name__ == '__main__':
    logging.basicConfig(level=os.environ.get('BUMBLE_LOGLEVEL', 'INFO').upper())
    asyncio.run(run_test_self())



# Copyright 2021-2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
from __future__ import annotations

import argparse
import asyncio
import sys

import bumble.logging
from bumble import core, l2cap
from bumble.device import Device
from bumble.transport import open_transport


# -----------------------------------------------------------------------------
async def main(
    config_file: str, transport: str, mode: int, peer_address: str, psm: int
) -> None:
    print('<<< connecting to HCI...')
    async with await open_transport(transport) as hci_transport:
        print('<<< connected')

        # Create a device
        device = Device.from_config_file_with_hci(
            config_file, hci_transport.source, hci_transport.sink
        )
        device.classic_enabled = True
        device.l2cap_channel_manager.extended_features.add(
            l2cap.L2CAP_Information_Request.ExtendedFeatures.ENHANCED_RETRANSMISSION_MODE
        )
        device.l2cap_channel_manager.extended_features.add(
            l2cap.L2CAP_Information_Request.ExtendedFeatures.FCS_OPTION
        )

        # Start the controller
        await device.power_on()

        # Start being discoverable and connectable
        await device.set_discoverable(True)
        await device.set_connectable(True)

        active_channel: l2cap.ClassicChannel | None = None

        def on_connection(channel: l2cap.ClassicChannel):
            def on_sdu(sdu: bytes):
                print(f'<<< {sdu.decode()}')

            channel.sink = on_sdu
            nonlocal active_channel
            active_channel = channel

        server = device.create_l2cap_server(
            spec=l2cap.ClassicChannelSpec(
                mode=l2cap.TransmissionMode(mode), psm=psm if psm else None
            ),
            handler=on_connection,
        )
        print(f'Listen L2CAP on channel {server.psm}')

        if peer_address:
            connection = await device.connect(
                peer_address, transport=core.PhysicalTransport.BR_EDR
            )
            channel = await connection.create_l2cap_channel(
                spec=l2cap.ClassicChannelSpec(
                    mode=l2cap.TransmissionMode(mode), psm=psm
                )
            )
            active_channel = channel

        while sdu := await asyncio.to_thread(lambda: input('>>> ')):
            if active_channel:
                active_channel.write(sdu.encode())

        await hci_transport.source.terminated


# -----------------------------------------------------------------------------
bumble.logging.setup_basic_logging('INFO')
parser = argparse.ArgumentParser()
parser.add_argument('config')
parser.add_argument('transport')
parser.add_argument('-p', '--peer_address', default='')
parser.add_argument(
    '-m', '--mode', default=l2cap.TransmissionMode.ENHANCED_RETRANSMISSION
)
parser.add_argument('--psm', default=0)
args = parser.parse_args(sys.argv[1:])
asyncio.run(main(args.config, args.transport, args.mode, args.peer_address, args.psm))



# Copyright 2021-2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import asyncio
from collections.abc import Callable, Iterable

import click

import bumble.logging
from bumble.colors import color
from bumble.core import ProtocolError
from bumble.device import Device, Peer
from bumble.gatt import Service
from bumble.profiles.battery_service import BatteryServiceProxy
from bumble.profiles.device_information_service import DeviceInformationServiceProxy
from bumble.profiles.gap import GenericAccessServiceProxy
from bumble.profiles.pacs import PublishedAudioCapabilitiesServiceProxy
from bumble.profiles.tmap import TelephonyAndMediaAudioServiceProxy
from bumble.profiles.vcs import VolumeControlServiceProxy
from bumble.transport import open_transport


# -----------------------------------------------------------------------------
async def try_show(function: Callable, *args, **kwargs) -> None:
    try:
        await function(*args, **kwargs)
    except ProtocolError as error:
        print(color('ERROR:', 'red'), error)


# -----------------------------------------------------------------------------
def show_services(services: Iterable[Service]) -> None:
    for service in services:
        print(color(str(service), 'cyan'))

        for characteristic in service.characteristics:
            print(color('  ' + str(characteristic), 'magenta'))


# -----------------------------------------------------------------------------
async def show_gap_information(
    gap_service: GenericAccessServiceProxy,
):
    print(color('### Generic Access Profile', 'yellow'))

    if gap_service.device_name:
        print(
            color(' Device Name:', 'green'),
            await gap_service.device_name.read_value(),
        )

    if gap_service.appearance:
        print(
            color(' Appearance: ', 'green'),
            await gap_service.appearance.read_value(),
        )

    print()


# -----------------------------------------------------------------------------
async def show_device_information(
    device_information_service: DeviceInformationServiceProxy,
):
    print(color('### Device Information', 'yellow'))

    if device_information_service.manufacturer_name:
        print(
            color('  Manufacturer Name:', 'green'),
            await device_information_service.manufacturer_name.read_value(),
        )

    if device_information_service.model_number:
        print(
            color('  Model Number:     ', 'green'),
            await device_information_service.model_number.read_value(),
        )

    if device_information_service.serial_number:
        print(
            color('  Serial Number:    ', 'green'),
            await device_information_service.serial_number.read_value(),
        )

    if device_information_service.firmware_revision:
        print(
            color('  Firmware Revision:', 'green'),
            await device_information_service.firmware_revision.read_value(),
        )

    print()


# -----------------------------------------------------------------------------
async def show_battery_level(
    battery_service: BatteryServiceProxy,
):
    print(color('### Battery Information', 'yellow'))

    if battery_service.battery_level:
        print(
            color('  Battery Level:', 'green'),
            await battery_service.battery_level.read_value(),
        )

    print()


# -----------------------------------------------------------------------------
async def show_tmas(
    tmas: TelephonyAndMediaAudioServiceProxy,
):
    print(color('### Telephony And Media Audio Service', 'yellow'))

    if tmas.role:
        role = await tmas.role.read_value()
        print(color('  Role:', 'green'), role)

    print()


# -----------------------------------------------------------------------------
async def show_pacs(pacs: PublishedAudioCapabilitiesServiceProxy) -> None:
    print(color('### Published Audio Capabilities Service', 'yellow'))

    contexts = await pacs.available_audio_contexts.read_value()
    print(color('  Available Audio Contexts:', 'green'), contexts)

    contexts = await pacs.supported_audio_contexts.read_value()
    print(color('  Supported Audio Contexts:', 'green'), contexts)

    if pacs.sink_pac:
        pac = await pacs.sink_pac.read_value()
        print(color('  Sink PAC:                ', 'green'), pac)

    if pacs.sink_audio_locations:
        audio_locations = await pacs.sink_audio_locations.read_value()
        print(color('  Sink Audio Locations:    ', 'green'), audio_locations)

    if pacs.source_pac:
        pac = await pacs.source_pac.read_value()
        print(color('  Source PAC:              ', 'green'), pac)

    if pacs.source_audio_locations:
        audio_locations = await pacs.source_audio_locations.read_value()
        print(color('  Source Audio Locations:  ', 'green'), audio_locations)

    print()


# -----------------------------------------------------------------------------
async def show_vcs(vcs: VolumeControlServiceProxy) -> None:
    print(color('### Volume Control Service', 'yellow'))

    volume_state = await vcs.volume_state.read_value()
    print(color('  Volume State:', 'green'), volume_state)

    volume_flags = await vcs.volume_flags.read_value()
    print(color('  Volume Flags:', 'green'), volume_flags)


# -----------------------------------------------------------------------------
async def show_device_info(peer, done: asyncio.Future | None) -> None:
    try:
        # Discover all services
        print(color('### Discovering Services and Characteristics', 'magenta'))
        await peer.discover_services()
        for service in peer.services:
            await service.discover_characteristics()

        print(color('=== Services ===', 'yellow'))
        show_services(peer.services)
        print()

        if gap_service := peer.create_service_proxy(GenericAccessServiceProxy):
            await try_show(show_gap_information, gap_service)

        if device_information_service := peer.create_service_proxy(
            DeviceInformationServiceProxy
        ):
            await try_show(show_device_information, device_information_service)

        if battery_service := peer.create_service_proxy(BatteryServiceProxy):
            await try_show(show_battery_level, battery_service)

        if tmas := peer.create_service_proxy(TelephonyAndMediaAudioServiceProxy):
            await try_show(show_tmas, tmas)

        if pacs := peer.create_service_proxy(PublishedAudioCapabilitiesServiceProxy):
            await try_show(show_pacs, pacs)

        if vcs := peer.create_service_proxy(VolumeControlServiceProxy):
            await try_show(show_vcs, vcs)

        if done is not None:
            done.set_result(None)
    except asyncio.CancelledError:
        print(color('!!! Operation canceled', 'red'))


# -----------------------------------------------------------------------------
async def async_main(device_config, encrypt, transport, address_or_name):
    async with await open_transport(transport) as (hci_source, hci_sink):
        # Create a device
        if device_config:
            device = Device.from_config_file_with_hci(
                device_config, hci_source, hci_sink
            )
        else:
            device = Device.with_hci(
                'Bumble', 'F0:F1:F2:F3:F4:F5', hci_source, hci_sink
            )
        await device.power_on()

        if address_or_name:
            # Connect to the target peer
            print(color('>>> Connecting...', 'green'))
            connection = await device.connect(address_or_name)
            print(color('>>> Connected', 'green'))

            # Encrypt the connection if required
            if encrypt:
                print(color('+++ Encrypting connection...', 'blue'))
                await connection.encrypt()
                print(color('+++ Encryption established', 'blue'))

            await show_device_info(Peer(connection), None)
        else:
            # Wait for a connection
            done = asyncio.get_running_loop().create_future()
            device.on(
                'connection',
                lambda connection: asyncio.create_task(
                    show_device_info(Peer(connection), done)
                ),
            )
            await device.start_advertising(auto_restart=True)

            print(color('### Waiting for connection...', 'blue'))
            await done


# -----------------------------------------------------------------------------
@click.command()
@click.option('--device-config', help='Device configuration', type=click.Path())
@click.option('--encrypt', help='Encrypt the connection', is_flag=True, default=False)
@click.argument('transport')
@click.argument('address-or-name', required=False)
def main(device_config, encrypt, transport, address_or_name):
    """
    Dump the GATT database on a remote device. If ADDRESS_OR_NAME is not specified,
    wait for an incoming connection.
    """
    bumble.logging.setup_basic_logging()
    asyncio.run(async_main(device_config, encrypt, transport, address_or_name))


# -----------------------------------------------------------------------------
if __name__ == '__main__':
    main()



# Copyright 2021-2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import asyncio
import logging
import os

import pytest

from bumble import device, gatt
from bumble.profiles import cap, csip

from .test_utils import TwoDevices

# -----------------------------------------------------------------------------
# Logging
# -----------------------------------------------------------------------------
logger = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
@pytest.mark.asyncio
async def test_cas():
    SIRK = bytes.fromhex('2f62c8ae41867d1bb619e788a2605faa')

    devices = TwoDevices()
    devices[0].add_service(
        cap.CommonAudioServiceService(
            csip.CoordinatedSetIdentificationService(
                set_identity_resolving_key=SIRK,
                set_identity_resolving_key_type=csip.SirkType.PLAINTEXT,
            )
        )
    )

    await devices.setup_connection()
    peer = device.Peer(devices.connections[1])
    cas_client = await peer.discover_service_and_create_proxy(
        cap.CommonAudioServiceServiceProxy
    )

    included_services = await peer.discover_included_services(cas_client.service_proxy)
    assert any(
        service.uuid == gatt.GATT_COORDINATED_SET_IDENTIFICATION_SERVICE
        for service in included_services
    )


# -----------------------------------------------------------------------------
async def run():
    await test_cas()


# -----------------------------------------------------------------------------
if __name__ == '__main__':
    logging.basicConfig(level=os.environ.get('BUMBLE_LOGLEVEL', 'INFO').upper())
    asyncio.run(run())



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import asyncio
import functools

from typing_extensions import Self

from bumble.controller import Controller
from bumble.device import Connection, Device
from bumble.hci import Address
from bumble.host import Host
from bumble.keys import PairingKeys
from bumble.link import LocalLink
from bumble.transport.common import AsyncPipeSink


# -----------------------------------------------------------------------------
class Devices:
    connections: dict[int, Connection]

    def __init__(self, num_devices: int) -> None:
        self.connections = {}

        self.link = LocalLink()
        addresses = [":".join([f"F{i}"] * 6) for i in range(num_devices)]
        self.controllers = [
            Controller(f'C{i + i}', link=self.link, public_address=addresses[i])
            for i in range(num_devices)
        ]
        self.devices = [
            Device(
                address=Address(addresses[i]),
                host=Host(self.controllers[i], AsyncPipeSink(self.controllers[i])),
            )
            for i in range(num_devices)
        ]

        for i in range(num_devices):
            self.devices[i].on(
                self.devices[i].EVENT_CONNECTION,
                functools.partial(self.on_connection, i),
            )

        self.paired = [
            asyncio.get_event_loop().create_future() for _ in range(num_devices)
        ]

    def on_connection(self, which: int, connection: Connection) -> None:
        self.connections[which] = connection
        connection.on(
            connection.EVENT_DISCONNECTION, lambda *_: self.on_disconnection(which)
        )

    def on_disconnection(self, which: int) -> None:
        self.connections.pop(which, None)

    def on_paired(self, which: int, keys: PairingKeys) -> None:
        self.paired[which].set_result(keys)

    async def setup_connection(self) -> None:
        # Start
        for dev in self.devices:
            await dev.power_on()

        # Connect devices
        for dev in self.devices[1:]:
            connection_future = asyncio.get_running_loop().create_future()
            dev.once(dev.EVENT_CONNECTION, connection_future.set_result)
            await dev.start_advertising(advertising_interval_min=1.0)
            await self.devices[0].connect(dev.random_address)
            await connection_future

    def __getitem__(self, index: int) -> Device:
        return self.devices[index]


# -----------------------------------------------------------------------------
class TwoDevices(Devices):
    def __init__(self) -> None:
        super().__init__(2)

    @classmethod
    async def create_with_connection(cls: type[Self]) -> Self:
        devices = cls()
        await devices.setup_connection()
        return devices


# -----------------------------------------------------------------------------
async def async_barrier():
    # TODO: Remove async barrier - this doesn't always mean what we want.
    for _ in range(3):
        ready = asyncio.get_running_loop().create_future()
        asyncio.get_running_loop().call_soon(ready.set_result, None)
        await ready



# Copyright 2021-2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import logging

import websockets.asyncio.server

from bumble.transport.common import ParserSource, PumpedPacketSink, Transport

# -----------------------------------------------------------------------------
# Logging
# -----------------------------------------------------------------------------
logger = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
async def open_ws_server_transport(spec: str) -> Transport:
    '''
    Open a WebSocket server transport.
    The parameter string has this syntax:
    <local-host>:<local-port>
    Where <local-host> may be the address of a local network interface, or '_'
    to accept connections on all local network interfaces.

    Example: _:9001
    '''

    class WsServerTransport(Transport):
        sink: PumpedPacketSink
        source: ParserSource
        connection: websockets.asyncio.server.ServerConnection | None
        server: websockets.asyncio.server.Server | None

        def __init__(self) -> None:
            source = ParserSource()
            sink = PumpedPacketSink(self.send_packet)
            self.connection = None
            self.server = None

            super().__init__(source, sink)

        async def serve(self, local_host: str, local_port: str) -> None:
            self.sink.start()
            # pylint: disable-next=no-member
            self.server = await websockets.asyncio.server.serve(
                handler=self.on_connection,
                host=local_host if local_host != '_' else None,
                port=int(local_port),
            )
            logger.debug(f'websocket server ready on port {local_port}')

        async def on_connection(
            self, connection: websockets.asyncio.server.ServerConnection
        ) -> None:
            logger.debug(
                f'new connection on {connection.local_address} '
                f'from {connection.remote_address}'
            )
            self.connection = connection
            # pylint: disable=no-member
            try:
                async for packet in connection:
                    if isinstance(packet, bytes):
                        self.source.parser.feed_data(packet)
                    else:
                        logger.warning('discarding packet: not a BINARY frame')
            except websockets.WebSocketException as error:
                logger.debug(f'exception while receiving packet: {error}')

            # We're now disconnected
            self.connection = None

        async def send_packet(self, packet: bytes) -> None:
            if self.connection is None:
                logger.debug('no connection, dropping packet')
                return
            await self.connection.send(packet)

    local_host, local_port = spec.rsplit(':', maxsplit=1)
    transport = WsServerTransport()
    await transport.serve(local_host, local_port)
    return transport



# Copyright 2021-2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import logging
import os
import struct
from collections.abc import Generator

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
from contextlib import contextmanager
from enum import IntEnum
from typing import BinaryIO

from bumble import core
from bumble.hci import HCI_COMMAND_PACKET, HCI_EVENT_PACKET

# -----------------------------------------------------------------------------
# Logging
# -----------------------------------------------------------------------------
logger = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
# Classes
# -----------------------------------------------------------------------------
class Snooper:
    """
    Base class for snooper implementations.

    A snooper is an object that will be provided with HCI packets as they are
    exchanged between a host and a controller.
    """

    class Direction(IntEnum):
        HOST_TO_CONTROLLER = 0
        CONTROLLER_TO_HOST = 1

    class DataLinkType(IntEnum):
        H1 = 1001
        H4 = 1002
        HCI_BSCP = 1003
        H5 = 1004

    def snoop(self, hci_packet: bytes, direction: Direction) -> None:
        """Snoop on an HCI packet."""


# -----------------------------------------------------------------------------
class BtSnooper(Snooper):
    """
    Snooper that saves HCI packets using the BTSnoop format, based on RFC 1761.
    """

    IDENTIFICATION_PATTERN = b'btsnoop\0'
    TIMESTAMP_ANCHOR = datetime.datetime(2000, 1, 1, tzinfo=datetime.timezone.utc)
    TIMESTAMP_DELTA = 0x00E03AB44A676000
    ONE_MS = datetime.timedelta(microseconds=1)

    def __init__(self, output: BinaryIO):
        self.output = output

        # Write the header
        self.output.write(
            self.IDENTIFICATION_PATTERN + struct.pack('>LL', 1, self.DataLinkType.H4)
        )

    def snoop(self, hci_packet: bytes, direction: Snooper.Direction) -> None:
        flags = int(direction)
        packet_type = hci_packet[0]
        if packet_type in (HCI_EVENT_PACKET, HCI_COMMAND_PACKET):
            flags |= 0x10

        # Compute the current timestamp
        timestamp = (
            int(
                (
                    datetime.datetime.now(tz=datetime.timezone.utc)
                    - self.TIMESTAMP_ANCHOR
                )
                / self.ONE_MS
            )
            + self.TIMESTAMP_DELTA
        )

        # Emit the record
        self.output.write(
            struct.pack(
                '>IIIIQ',
                len(hci_packet),  # Original Length
                len(hci_packet),  # Included Length
                flags,  # Packet Flags
                0,  # Cumulative Drops
                timestamp,  # Timestamp
            )
            + hci_packet
        )


# -----------------------------------------------------------------------------
class PcapSnooper(Snooper):
    """
    Snooper that saves or streames HCI packets using the PCAP format.
    """

    PCAP_MAGIC = 0xA1B2C3D4
    DLT_BLUETOOTH_HCI_H4_WITH_PHDR = 201

    def __init__(self, output: BinaryIO):
        self.output = output

        # Write the header
        self.output.write(
            struct.pack(
                "<IHHIIII",
                self.PCAP_MAGIC,
                2,  # Major PCAP Version
                4,  # Minor PCAP Version
                0,  # Reserved 1
                0,  # Reserved 2
                65535,  # SnapLen
                # FCS and f are set to 0 implicitly by the next line
                self.DLT_BLUETOOTH_HCI_H4_WITH_PHDR,  # The DLT in this PCAP
            )
        )

    def snoop(self, hci_packet: bytes, direction: Snooper.Direction):
        now = datetime.datetime.now(datetime.timezone.utc)
        sec = int(now.timestamp())
        usec = now.microsecond

        # Emit the record
        self.output.write(
            struct.pack(
                "<IIII",
                sec,  # Timestamp (Seconds)
                usec,  # Timestamp (Microseconds)
                len(hci_packet) + 4,
                len(hci_packet) + 4,  # +4 because of the addtional direction info...
            )
            + struct.pack(">I", int(direction))  # ...thats being added here
            + hci_packet
        )
        self.output.flush()  # flush after every packet for live logging


# -----------------------------------------------------------------------------
_SNOOPER_INSTANCE_COUNT = 0


@contextmanager
def create_snooper(spec: str) -> Generator[Snooper, None, None]:
    """
    Create a snooper given a specification string.

    The general syntax for the specification string is:
      <snooper-type>:<type-specific-arguments>

    Supported snooper types are:

      btsnoop
        The syntax for the type-specific arguments for this type is:
        <io-type>:<io-type-specific-arguments>

        Supported I/O types are:

        file
          The type-specific arguments for this I/O type is a string that is converted
          to a file path using the python `str.format()` string formatting. The log
          records will be written to that file if it can be opened/created.
          The keyword args that may be referenced by the string pattern are:
            now: the value of `datetime.now()`
            utcnow: the value of `datetime.now(tz=datetime.timezone.utc)`
            pid: the current process ID.
            instance: the instance ID in the current process.

      pcapsnoop
        The syntax for the type-specific arguments for this type is:
        <io-type>:<io-type-specific-arguments>

        Supported I/O types are:

        file
          The type-specific arguments for this I/O type is a string that is converted
          to a file path using the python `str.format()` string formatting. The log
          records will be written to that file if it can be opened/created.
          The keyword args that may be referenced by the string pattern are:
            now: the value of `datetime.now()`
            utcnow: the value of `datetime.now(tz=datetime.timezone.utc)`
            pid: the current process ID.
            instance: the instance ID in the current process.

        pipe
          The type-specific arguments for this I/O type is a string that is converted
          to a path using the python `str.format()` string formatting. The log
          records will be written to the named pipe referenced by this path
          if it can be opened. The keyword args that may be referenced by the
          string pattern are:
            now: the value of `datetime.now()`
            utcnow: the value of `datetime.now(tz=datetime.timezone.utc)`
            pid: the current process ID.
            instance: the instance ID in the current process.

    Examples:
      btsnoop:file:my_btsnoop.log
      btsnoop:file:/tmp/bumble_{now:%Y-%m-%d-%H:%M:%S}_{pid}.log
      pcapsnoop:pipe:/tmp/bumble-extcap


    """
    if ':' not in spec:
        raise core.InvalidArgumentError('snooper type prefix missing')

    snooper_type, snooper_args = spec.split(':', maxsplit=1)

    global _SNOOPER_INSTANCE_COUNT

    if snooper_type == 'btsnoop':
        if ':' not in snooper_args:
            raise core.InvalidArgumentError('I/O type for btsnoop snooper type missing')

        io_type, io_name = snooper_args.split(':', maxsplit=1)
        if io_type == 'file':
            # Process the file name string pattern.
            file_path = io_name.format(
                now=datetime.datetime.now(),
                utcnow=datetime.datetime.now(tz=datetime.timezone.utc),
                pid=os.getpid(),
                instance=_SNOOPER_INSTANCE_COUNT,
            )

            # Open the file
            logger.debug(f'Snoop file: {file_path}')
            with open(file_path, 'wb') as snoop_file:
                _SNOOPER_INSTANCE_COUNT += 1
                yield BtSnooper(snoop_file)
                _SNOOPER_INSTANCE_COUNT -= 1
                return

    elif snooper_type == 'pcapsnoop':
        if ':' not in snooper_args:
            raise core.InvalidArgumentError(
                'I/O type for pcapsnoop snooper type missing'
            )

        io_type, io_name = snooper_args.split(':', maxsplit=1)
        if io_type in {'pipe', 'file'}:
            # Process the file name string pattern.
            file_path = io_name.format(
                now=datetime.datetime.now(),
                utcnow=datetime.datetime.now(tz=datetime.timezone.utc),
                pid=os.getpid(),
                instance=_SNOOPER_INSTANCE_COUNT,
            )

            # Open a file or pipe
            logger.debug(f'PCAP file: {file_path}')

            # Pipes we have to open with unbuffered binary I/O
            # so we pass ``buffering`` for pipes but not for files
            pcap_file: BinaryIO
            if io_type == 'pipe':
                pcap_file = open(file_path, 'wb', buffering=0)
            else:
                pcap_file = open(file_path, 'wb')

            with pcap_file:
                _SNOOPER_INSTANCE_COUNT += 1
                yield PcapSnooper(pcap_file)
                _SNOOPER_INSTANCE_COUNT -= 1
                return

        raise core.InvalidArgumentError(f'I/O type {io_type} not supported')

    raise core.InvalidArgumentError(f'snooper type {snooper_type} not found')



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: emulated_bluetooth.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import emulated_bluetooth_packets_pb2 as emulated__bluetooth__packets__pb2
from . import emulated_bluetooth_device_pb2 as emulated__bluetooth__device__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x18\x65mulated_bluetooth.proto\x12\x1b\x61ndroid.emulation.bluetooth\x1a emulated_bluetooth_packets.proto\x1a\x1f\x65mulated_bluetooth_device.proto\"\x19\n\x07RawData\x12\x0e\n\x06packet\x18\x01 \x01(\x0c\"a\n\x12RegistrationStatus\x12K\n\x12\x63\x61llback_device_id\x18\x01 \x01(\x0b\x32/.android.emulation.bluetooth.CallbackIdentifier2\xbb\x03\n\x18\x45mulatedBluetoothService\x12\x64\n\x12registerClassicPhy\x12$.android.emulation.bluetooth.RawData\x1a$.android.emulation.bluetooth.RawData(\x01\x30\x01\x12`\n\x0eregisterBlePhy\x12$.android.emulation.bluetooth.RawData\x1a$.android.emulation.bluetooth.RawData(\x01\x30\x01\x12g\n\x11registerHCIDevice\x12&.android.emulation.bluetooth.HCIPacket\x1a&.android.emulation.bluetooth.HCIPacket(\x01\x30\x01\x12n\n\x12registerGattDevice\x12\'.android.emulation.bluetooth.GattDevice\x1a/.android.emulation.bluetooth.RegistrationStatusB\"\n\x1e\x63om.android.emulator.bluetoothP\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'emulated_bluetooth_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\036com.android.emulator.bluetoothP\001'
  _RAWDATA._serialized_start=124
  _RAWDATA._serialized_end=149
  _REGISTRATIONSTATUS._serialized_start=151
  _REGISTRATIONSTATUS._serialized_end=248
  _EMULATEDBLUETOOTHSERVICE._serialized_start=251
  _EMULATEDBLUETOOTHSERVICE._serialized_end=694
# @@protoc_insertion_point(module_scope)



# Copyright 2021-2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
from __future__ import annotations

import pytest

from bumble import device, gatt
from bumble.profiles import gatt_service

from . import test_utils


# -----------------------------------------------------------------------------
@pytest.mark.asyncio
async def test_database_hash():
    devices = await test_utils.TwoDevices.create_with_connection()
    devices[0].gatt_server.services.clear()
    devices[0].gatt_server.attributes.clear()
    devices[0].gatt_server.attributes_by_handle.clear()
    devices[0].add_service(
        gatt.Service(
            gatt.GATT_GENERIC_ACCESS_SERVICE,
            characteristics=[
                gatt.Characteristic(
                    gatt.GATT_DEVICE_NAME_CHARACTERISTIC,
                    (
                        gatt.Characteristic.Properties.READ
                        | gatt.Characteristic.Properties.WRITE
                    ),
                    gatt.Characteristic.Permissions.READ_REQUIRES_AUTHENTICATION,
                ),
                gatt.Characteristic(
                    gatt.GATT_APPEARANCE_CHARACTERISTIC,
                    gatt.Characteristic.Properties.READ,
                    gatt.Characteristic.Permissions.READ_REQUIRES_AUTHENTICATION,
                ),
            ],
        )
    )
    devices[0].add_service(
        gatt_service.GenericAttributeProfileService(
            server_supported_features=None,
            database_hash_enabled=True,
            service_change_enabled=True,
        )
    )
    devices[0].gatt_server.add_attribute(
        gatt.Service(gatt.GATT_GLUCOSE_SERVICE, characteristics=[])
    )
    # There is a special attribute order in the spec, so we need to add attribute one by
    # one here.
    battery_service = gatt.Service(
        gatt.GATT_BATTERY_SERVICE,
        characteristics=[
            gatt.Characteristic(
                gatt.GATT_BATTERY_LEVEL_CHARACTERISTIC,
                properties=gatt.Characteristic.Properties.READ,
                permissions=gatt.Characteristic.Permissions.READ_REQUIRES_AUTHENTICATION,
            )
        ],
        primary=False,
    )
    battery_service.handle = 0x0014
    battery_service.end_group_handle = 0x0016
    devices[0].gatt_server.add_attribute(
        gatt.IncludedServiceDeclaration(battery_service)
    )
    c = gatt.Characteristic(
        '2A18',
        properties=(
            gatt.Characteristic.Properties.READ
            | gatt.Characteristic.Properties.INDICATE
            | gatt.Characteristic.Properties.EXTENDED_PROPERTIES
        ),
        permissions=gatt.Characteristic.Permissions.READ_REQUIRES_AUTHENTICATION,
    )
    devices[0].gatt_server.add_attribute(
        gatt.CharacteristicDeclaration(c, devices[0].gatt_server.next_handle() + 1)
    )
    devices[0].gatt_server.add_attribute(c)
    devices[0].gatt_server.add_attribute(
        gatt.Descriptor(
            gatt.GATT_CLIENT_CHARACTERISTIC_CONFIGURATION_DESCRIPTOR,
            gatt.Descriptor.Permissions.READ_REQUIRES_AUTHENTICATION,
            b'\x02\x00',
        ),
    )
    devices[0].gatt_server.add_attribute(
        gatt.Descriptor(
            gatt.GATT_CHARACTERISTIC_EXTENDED_PROPERTIES_DESCRIPTOR,
            gatt.Descriptor.Permissions.READ_REQUIRES_AUTHENTICATION,
            b'\x00\x00',
        ),
    )
    devices[0].add_service(battery_service)

    peer = device.Peer(devices.connections[1])
    client = await peer.discover_service_and_create_proxy(
        gatt_service.GenericAttributeProfileServiceProxy
    )
    assert client.database_hash_characteristic
    assert await client.database_hash_characteristic.read_value() == bytes.fromhex(
        'F1CA2D48ECF58BAC8A8830BBB9FBA990'
    )


# -----------------------------------------------------------------------------
@pytest.mark.asyncio
async def test_service_changed():
    devices = await test_utils.TwoDevices.create_with_connection()
    assert (service := devices[0].gatt_service)

    peer = device.Peer(devices.connections[1])
    assert (
        client := await peer.discover_service_and_create_proxy(
            gatt_service.GenericAttributeProfileServiceProxy
        )
    )
    assert client.service_changed_characteristic
    indications = []
    await client.service_changed_characteristic.subscribe(
        indications.append, prefer_notify=False
    )
    await devices[0].indicate_subscribers(
        service.service_changed_characteristic, b'1234'
    )
    await test_utils.async_barrier()
    assert indications[0] == b'1234'



from bumble.transport.grpc_protobuf.netsim import common_pb2 as _common_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from bumble.transport.grpc_protobuf.rootcanal import configuration_pb2 as _configuration_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class PhyKind(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    NONE: _ClassVar[PhyKind]
    BLUETOOTH_CLASSIC: _ClassVar[PhyKind]
    BLUETOOTH_LOW_ENERGY: _ClassVar[PhyKind]
    WIFI: _ClassVar[PhyKind]
    UWB: _ClassVar[PhyKind]
    WIFI_RTT: _ClassVar[PhyKind]
NONE: PhyKind
BLUETOOTH_CLASSIC: PhyKind
BLUETOOTH_LOW_ENERGY: PhyKind
WIFI: PhyKind
UWB: PhyKind
WIFI_RTT: PhyKind

class Position(_message.Message):
    __slots__ = ("x", "y", "z")
    X_FIELD_NUMBER: _ClassVar[int]
    Y_FIELD_NUMBER: _ClassVar[int]
    Z_FIELD_NUMBER: _ClassVar[int]
    x: float
    y: float
    z: float
    def __init__(self, x: _Optional[float] = ..., y: _Optional[float] = ..., z: _Optional[float] = ...) -> None: ...

class Orientation(_message.Message):
    __slots__ = ("yaw", "pitch", "roll")
    YAW_FIELD_NUMBER: _ClassVar[int]
    PITCH_FIELD_NUMBER: _ClassVar[int]
    ROLL_FIELD_NUMBER: _ClassVar[int]
    yaw: float
    pitch: float
    roll: float
    def __init__(self, yaw: _Optional[float] = ..., pitch: _Optional[float] = ..., roll: _Optional[float] = ...) -> None: ...

class Chip(_message.Message):
    __slots__ = ("kind", "id", "name", "manufacturer", "product_name", "bt", "ble_beacon", "uwb", "wifi", "offset")
    class Radio(_message.Message):
        __slots__ = ("state", "range", "tx_count", "rx_count")
        STATE_FIELD_NUMBER: _ClassVar[int]
        RANGE_FIELD_NUMBER: _ClassVar[int]
        TX_COUNT_FIELD_NUMBER: _ClassVar[int]
        RX_COUNT_FIELD_NUMBER: _ClassVar[int]
        state: bool
        range: float
        tx_count: int
        rx_count: int
        def __init__(self, state: bool = ..., range: _Optional[float] = ..., tx_count: _Optional[int] = ..., rx_count: _Optional[int] = ...) -> None: ...
    class Bluetooth(_message.Message):
        __slots__ = ("low_energy", "classic", "address", "bt_properties")
        LOW_ENERGY_FIELD_NUMBER: _ClassVar[int]
        CLASSIC_FIELD_NUMBER: _ClassVar[int]
        ADDRESS_FIELD_NUMBER: _ClassVar[int]
        BT_PROPERTIES_FIELD_NUMBER: _ClassVar[int]
        low_energy: Chip.Radio
        classic: Chip.Radio
        address: str
        bt_properties: _configuration_pb2.Controller
        def __init__(self, low_energy: _Optional[_Union[Chip.Radio, _Mapping]] = ..., classic: _Optional[_Union[Chip.Radio, _Mapping]] = ..., address: _Optional[str] = ..., bt_properties: _Optional[_Union[_configuration_pb2.Controller, _Mapping]] = ...) -> None: ...
    class BleBeacon(_message.Message):
        __slots__ = ("bt", "address", "settings", "adv_data", "scan_response")
        class AdvertiseSettings(_message.Message):
            __slots__ = ("advertise_mode", "milliseconds", "tx_power_level", "dbm", "scannable", "timeout")
            class AdvertiseMode(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
                __slots__ = ()
                LOW_POWER: _ClassVar[Chip.BleBeacon.AdvertiseSettings.AdvertiseMode]
                BALANCED: _ClassVar[Chip.BleBeacon.AdvertiseSettings.AdvertiseMode]
                LOW_LATENCY: _ClassVar[Chip.BleBeacon.AdvertiseSettings.AdvertiseMode]
            LOW_POWER: Chip.BleBeacon.AdvertiseSettings.AdvertiseMode
            BALANCED: Chip.BleBeacon.AdvertiseSettings.AdvertiseMode
            LOW_LATENCY: Chip.BleBeacon.AdvertiseSettings.AdvertiseMode
            class AdvertiseTxPower(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
                __slots__ = ()
                ULTRA_LOW: _ClassVar[Chip.BleBeacon.AdvertiseSettings.AdvertiseTxPower]
                LOW: _ClassVar[Chip.BleBeacon.AdvertiseSettings.AdvertiseTxPower]
                MEDIUM: _ClassVar[Chip.BleBeacon.AdvertiseSettings.AdvertiseTxPower]
                HIGH: _ClassVar[Chip.BleBeacon.AdvertiseSettings.AdvertiseTxPower]
            ULTRA_LOW: Chip.BleBeacon.AdvertiseSettings.AdvertiseTxPower
            LOW: Chip.BleBeacon.AdvertiseSettings.AdvertiseTxPower
            MEDIUM: Chip.BleBeacon.AdvertiseSettings.AdvertiseTxPower
            HIGH: Chip.BleBeacon.AdvertiseSettings.AdvertiseTxPower
            ADVERTISE_MODE_FIELD_NUMBER: _ClassVar[int]
            MILLISECONDS_FIELD_NUMBER: _ClassVar[int]
            TX_POWER_LEVEL_FIELD_NUMBER: _ClassVar[int]
            DBM_FIELD_NUMBER: _ClassVar[int]
            SCANNABLE_FIELD_NUMBER: _ClassVar[int]
            TIMEOUT_FIELD_NUMBER: _ClassVar[int]
            advertise_mode: Chip.BleBeacon.AdvertiseSettings.AdvertiseMode
            milliseconds: int
            tx_power_level: Chip.BleBeacon.AdvertiseSettings.AdvertiseTxPower
            dbm: int
            scannable: bool
            timeout: int
            def __init__(self, advertise_mode: _Optional[_Union[Chip.BleBeacon.AdvertiseSettings.AdvertiseMode, str]] = ..., milliseconds: _Optional[int] = ..., tx_power_level: _Optional[_Union[Chip.BleBeacon.AdvertiseSettings.AdvertiseTxPower, str]] = ..., dbm: _Optional[int] = ..., scannable: bool = ..., timeout: _Optional[int] = ...) -> None: ...
        class AdvertiseData(_message.Message):
            __slots__ = ("include_device_name", "include_tx_power_level", "manufacturer_data", "services")
            class Service(_message.Message):
                __slots__ = ("uuid", "data")
                UUID_FIELD_NUMBER: _ClassVar[int]
                DATA_FIELD_NUMBER: _ClassVar[int]
                uuid: str
                data: bytes
                def __init__(self, uuid: _Optional[str] = ..., data: _Optional[bytes] = ...) -> None: ...
            INCLUDE_DEVICE_NAME_FIELD_NUMBER: _ClassVar[int]
            INCLUDE_TX_POWER_LEVEL_FIELD_NUMBER: _ClassVar[int]
            MANUFACTURER_DATA_FIELD_NUMBER: _ClassVar[int]
            SERVICES_FIELD_NUMBER: _ClassVar[int]
            include_device_name: bool
            include_tx_power_level: bool
            manufacturer_data: bytes
            services: _containers.RepeatedCompositeFieldContainer[Chip.BleBeacon.AdvertiseData.Service]
            def __init__(self, include_device_name: bool = ..., include_tx_power_level: bool = ..., manufacturer_data: _Optional[bytes] = ..., services: _Optional[_Iterable[_Union[Chip.BleBeacon.AdvertiseData.Service, _Mapping]]] = ...) -> None: ...
        BT_FIELD_NUMBER: _ClassVar[int]
        ADDRESS_FIELD_NUMBER: _ClassVar[int]
        SETTINGS_FIELD_NUMBER: _ClassVar[int]
        ADV_DATA_FIELD_NUMBER: _ClassVar[int]
        SCAN_RESPONSE_FIELD_NUMBER: _ClassVar[int]
        bt: Chip.Bluetooth
        address: str
        settings: Chip.BleBeacon.AdvertiseSettings
        adv_data: Chip.BleBeacon.AdvertiseData
        scan_response: Chip.BleBeacon.AdvertiseData
        def __init__(self, bt: _Optional[_Union[Chip.Bluetooth, _Mapping]] = ..., address: _Optional[str] = ..., settings: _Optional[_Union[Chip.BleBeacon.AdvertiseSettings, _Mapping]] = ..., adv_data: _Optional[_Union[Chip.BleBeacon.AdvertiseData, _Mapping]] = ..., scan_response: _Optional[_Union[Chip.BleBeacon.AdvertiseData, _Mapping]] = ...) -> None: ...
    KIND_FIELD_NUMBER: _ClassVar[int]
    ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    MANUFACTURER_FIELD_NUMBER: _ClassVar[int]
    PRODUCT_NAME_FIELD_NUMBER: _ClassVar[int]
    BT_FIELD_NUMBER: _ClassVar[int]
    BLE_BEACON_FIELD_NUMBER: _ClassVar[int]
    UWB_FIELD_NUMBER: _ClassVar[int]
    WIFI_FIELD_NUMBER: _ClassVar[int]
    OFFSET_FIELD_NUMBER: _ClassVar[int]
    kind: _common_pb2.ChipKind
    id: int
    name: str
    manufacturer: str
    product_name: str
    bt: Chip.Bluetooth
    ble_beacon: Chip.BleBeacon
    uwb: Chip.Radio
    wifi: Chip.Radio
    offset: Position
    def __init__(self, kind: _Optional[_Union[_common_pb2.ChipKind, str]] = ..., id: _Optional[int] = ..., name: _Optional[str] = ..., manufacturer: _Optional[str] = ..., product_name: _Optional[str] = ..., bt: _Optional[_Union[Chip.Bluetooth, _Mapping]] = ..., ble_beacon: _Optional[_Union[Chip.BleBeacon, _Mapping]] = ..., uwb: _Optional[_Union[Chip.Radio, _Mapping]] = ..., wifi: _Optional[_Union[Chip.Radio, _Mapping]] = ..., offset: _Optional[_Union[Position, _Mapping]] = ...) -> None: ...

class ChipCreate(_message.Message):
    __slots__ = ("kind", "address", "name", "manufacturer", "product_name", "ble_beacon", "bt_properties")
    class BleBeaconCreate(_message.Message):
        __slots__ = ("address", "settings", "adv_data", "scan_response")
        ADDRESS_FIELD_NUMBER: _ClassVar[int]
        SETTINGS_FIELD_NUMBER: _ClassVar[int]
        ADV_DATA_FIELD_NUMBER: _ClassVar[int]
        SCAN_RESPONSE_FIELD_NUMBER: _ClassVar[int]
        address: str
        settings: Chip.BleBeacon.AdvertiseSettings
        adv_data: Chip.BleBeacon.AdvertiseData
        scan_response: Chip.BleBeacon.AdvertiseData
        def __init__(self, address: _Optional[str] = ..., settings: _Optional[_Union[Chip.BleBeacon.AdvertiseSettings, _Mapping]] = ..., adv_data: _Optional[_Union[Chip.BleBeacon.AdvertiseData, _Mapping]] = ..., scan_response: _Optional[_Union[Chip.BleBeacon.AdvertiseData, _Mapping]] = ...) -> None: ...
    KIND_FIELD_NUMBER: _ClassVar[int]
    ADDRESS_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    MANUFACTURER_FIELD_NUMBER: _ClassVar[int]
    PRODUCT_NAME_FIELD_NUMBER: _ClassVar[int]
    BLE_BEACON_FIELD_NUMBER: _ClassVar[int]
    BT_PROPERTIES_FIELD_NUMBER: _ClassVar[int]
    kind: _common_pb2.ChipKind
    address: str
    name: str
    manufacturer: str
    product_name: str
    ble_beacon: ChipCreate.BleBeaconCreate
    bt_properties: _configuration_pb2.Controller
    def __init__(self, kind: _Optional[_Union[_common_pb2.ChipKind, str]] = ..., address: _Optional[str] = ..., name: _Optional[str] = ..., manufacturer: _Optional[str] = ..., product_name: _Optional[str] = ..., ble_beacon: _Optional[_Union[ChipCreate.BleBeaconCreate, _Mapping]] = ..., bt_properties: _Optional[_Union[_configuration_pb2.Controller, _Mapping]] = ...) -> None: ...

class Device(_message.Message):
    __slots__ = ("id", "name", "visible", "position", "orientation", "chips")
    ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    VISIBLE_FIELD_NUMBER: _ClassVar[int]
    POSITION_FIELD_NUMBER: _ClassVar[int]
    ORIENTATION_FIELD_NUMBER: _ClassVar[int]
    CHIPS_FIELD_NUMBER: _ClassVar[int]
    id: int
    name: str
    visible: bool
    position: Position
    orientation: Orientation
    chips: _containers.RepeatedCompositeFieldContainer[Chip]
    def __init__(self, id: _Optional[int] = ..., name: _Optional[str] = ..., visible: bool = ..., position: _Optional[_Union[Position, _Mapping]] = ..., orientation: _Optional[_Union[Orientation, _Mapping]] = ..., chips: _Optional[_Iterable[_Union[Chip, _Mapping]]] = ...) -> None: ...

class DeviceCreate(_message.Message):
    __slots__ = ("name", "position", "orientation", "chips")
    NAME_FIELD_NUMBER: _ClassVar[int]
    POSITION_FIELD_NUMBER: _ClassVar[int]
    ORIENTATION_FIELD_NUMBER: _ClassVar[int]
    CHIPS_FIELD_NUMBER: _ClassVar[int]
    name: str
    position: Position
    orientation: Orientation
    chips: _containers.RepeatedCompositeFieldContainer[ChipCreate]
    def __init__(self, name: _Optional[str] = ..., position: _Optional[_Union[Position, _Mapping]] = ..., orientation: _Optional[_Union[Orientation, _Mapping]] = ..., chips: _Optional[_Iterable[_Union[ChipCreate, _Mapping]]] = ...) -> None: ...

class Scene(_message.Message):
    __slots__ = ("devices",)
    DEVICES_FIELD_NUMBER: _ClassVar[int]
    devices: _containers.RepeatedCompositeFieldContainer[Device]
    def __init__(self, devices: _Optional[_Iterable[_Union[Device, _Mapping]]] = ...) -> None: ...

class Capture(_message.Message):
    __slots__ = ("id", "chip_kind", "device_name", "state", "size", "records", "timestamp", "valid")
    ID_FIELD_NUMBER: _ClassVar[int]
    CHIP_KIND_FIELD_NUMBER: _ClassVar[int]
    DEVICE_NAME_FIELD_NUMBER: _ClassVar[int]
    STATE_FIELD_NUMBER: _ClassVar[int]
    SIZE_FIELD_NUMBER: _ClassVar[int]
    RECORDS_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    VALID_FIELD_NUMBER: _ClassVar[int]
    id: int
    chip_kind: _common_pb2.ChipKind
    device_name: str
    state: bool
    size: int
    records: int
    timestamp: _timestamp_pb2.Timestamp
    valid: bool
    def __init__(self, id: _Optional[int] = ..., chip_kind: _Optional[_Union[_common_pb2.ChipKind, str]] = ..., device_name: _Optional[str] = ..., state: bool = ..., size: _Optional[int] = ..., records: _Optional[int] = ..., timestamp: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., valid: bool = ...) -> None: ...
